// TBehaviorNetwork.cpp: implementation of the TBehaviorNetwork class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "NetworkHeader.h"
#include "TNetworkView.h"
#include "TRuleView.h"
#include "TGoal.h"
#include "TAct.h"
#include "TPercept.h"
#include "TAction.h"
#include "TNetworkParams.h"
#include "TScreenParamsNetwork.h"
#include "TCompetence.h"
#include "TParseString.h"
#include "TProposition.h"
#include "TWorldState.h"
#include "TBehaviorNetwork.h"

#include <time.h>


#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

TBehaviorNetwork::TBehaviorNetwork(TNetworkView* pNView, TRuleView* pRView)
{
	// create a list of modules and add the percept and act object
	m_pRulesView = pRView;
	m_pNetworkView = pNView;
	m_pGoalList = new CObList();
	m_pCompetenceList = new CObList();
	m_pNetworkParams = new TNetworkParams();
	m_pAct = new TAct();
	m_pPercept = new TPercept();

	// for blocksworld
	m_pObjects = new CStringList();
}

void TBehaviorNetwork::Initialize()
{
	m_pLastAction = NULL;
	m_lNoOfActions = 0L;
	m_lNoOfActionChanges = 0L;
	m_dAvgNoOfExecutableModules = 0.0;
	m_dThreshold = m_pNetworkParams->GetTheta();

	// seed random generator (only needed for random action selection)
	srand((unsigned) time(NULL));
		
	// only needed for blocksworld
	m_bBlocksworld = FALSE;
}
	
TBehaviorNetwork::~TBehaviorNetwork()
{
	DeleteNetwork();

	delete m_pAct;
	delete m_pPercept;
	delete m_pCompetenceList;
	delete m_pNetworkParams;
	delete m_pGoalList;
	// for blocksworld
	delete m_pObjects;
}

void TBehaviorNetwork::DeleteNetwork()
{
	// delete the goals
	while (!m_pGoalList->IsEmpty())
		delete m_pGoalList->RemoveHead();

	// delete competences
	while (!m_pCompetenceList->IsEmpty())
		delete m_pCompetenceList->RemoveHead();

	// set selected module to null
	m_pNetworkView->SetSelectedModule(NULL);

	// for blocksworld
	// delete all objects names
	m_pObjects->RemoveAll();
}


/*-----------------------------------------------------------------------------
Function: double TBehaviorNetwork::DoActivationSpreading()
Date    : 19.03.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: sum of activation in the network
Description: calculates the new input to the network and the spreading of activation
				 within the network. Only one cycle of activation spreading is executed.
				 It is assumed, that there is real perception available, that is, the
				 agent is either connected to the server, or provided with some logged
				 or fake perception
-----------------------------------------------------------------------------*/
double TBehaviorNetwork::DoActivationSpreading()
{
	POSITION			pos;
	TCompetence*	module;
	double			activity;

	// first calculate all executabilities (and initialize sumtaup)
	pos = m_pCompetenceList->GetHeadPosition();
	while (pos != NULL)
	{
		module = (TCompetence*) m_pCompetenceList->GetNext(pos);
		module->CalculateExecutability();
	}

	// calculate the activation of competences from goals and perceptions
	pos = m_pCompetenceList->GetHeadPosition();
	while (pos != NULL)
	{
		module = (TCompetence*) m_pCompetenceList->GetNext(pos);
		module->CalculateActivation(GetNoOfRules(), GetActivationSum());
	}

	// calculate the activation spreading of competences
	CalculateSpreading();

	// calculate and set are separated, so that new activations depend all on the 
	// current activations of modules and not of their new energy
	activity = 0.0;
	pos = m_pCompetenceList->GetHeadPosition();
	while (pos != NULL)
		activity += ((TCompetence*) m_pCompetenceList->GetNext(pos))->SetToNewActivation();

	// draw the network
	if (!m_pNetworkView->GetDrawAtEndOfActionSelection())
		m_pNetworkView->RedrawView();

	return activity;
}

/*-----------------------------------------------------------------------------
Function: void TBehaviorNetwork::DoActionSelection()
Date    : 08.01.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: first : calculates the activation of each module
				 second: sets the new activation of each module and determines the 
							module(s) with highest activation
				 third : executes the module(s) with highest activation, if it is 
							higher than the threshold
-----------------------------------------------------------------------------*/
BOOL TBehaviorNetwork::DoActionSelection()
{
	POSITION			pos;
	TCompetence*	module;
	int				iNoOfExecutableModules = 0;
	int				iNoOfExecutedModules = 0;
	CObList			listOfOrderedCompetences;

	pos = m_pCompetenceList->GetHeadPosition();
	while (pos != NULL)
	{
		module = (TCompetence*) m_pCompetenceList->GetNext(pos);
		module->SetExecuted(FALSE);

		if (module->GetExecutability() > 0.01)
		{
			// sort the executable modules by activation and executability
			InsertSorted(listOfOrderedCompetences, module);

			// for statistics
			iNoOfExecutableModules++;
		}
	}

	// check if highest module is higher than threshold 
	pos = listOfOrderedCompetences.GetHeadPosition();
	while (pos != NULL) 
	{
		module = (TCompetence*) listOfOrderedCompetences.GetNext(pos);
		if (module->CheckForExecution(m_dThreshold, m_bBlocksworld))
		{
			// only execute one module is standard
			iNoOfExecutedModules++;
			pos = NULL;
		}
	}

	if (iNoOfExecutedModules == 0)
		return FALSE;
	else
	{
		// at least one action has been executed, reset theta
		m_dThreshold = m_pNetworkParams->GetTheta();

		if (!m_bBlocksworld)
		{
			// statistics
			m_dAvgNoOfExecutableModules = (m_dAvgNoOfExecutableModules * m_lNoOfActions 
													+ iNoOfExecutableModules) / (double)(m_lNoOfActions + 1L);
			m_lNoOfActions++;
			if (m_pLastAction != module->GetAction())
				m_lNoOfActionChanges++;
			m_pLastAction = module->GetAction();
			if (m_lNoOfActions < MAX_CYCLES)
				m_aiActionExecuted[m_lNoOfActions] = m_pLastAction->GetIndex();
		}		
		return TRUE;
	}
}

/*
// this is the version, which does random action selection of the executable modules
BOOL TBehaviorNetwork::DoActionSelection()
{
	POSITION			pos;
	CObList			executableModules;
	TCompetence*	module;
	int				iIndex;

	pos = m_pCompetenceList->GetHeadPosition();
	while (pos != NULL)
	{
		module = (TCompetence*) m_pCompetenceList->GetNext(pos);
		module->SetExecuted(FALSE);

		// find executable modules
		if (module->GetExecutability() > 0.01)
			executableModules.AddTail(module);
	}

	// if there are no modules
	if (executableModules.GetCount() == 0)
		return TRUE;

	// execute the action
	iIndex = rand() / (RAND_MAX / executableModules.GetCount());
	pos = executableModules.FindIndex(iIndex);
	module = (TCompetence*) executableModules.GetAt(pos);
	module->ExecuteModule(1.0);
	m_pLastAction = module->GetAction();
	return TRUE;
}
*/

/*-----------------------------------------------------------------------------
Function: void TBehaviorNetwork::InsertSorted(CObList& list, TCompetence* module)
Date    : 08.03.99
-------------------------------------------------------------------------------
Parameter: reference to the list to insert
			  module to insert
Returnval: (void)
Description: inserts the module into the list sorted by the activationexecutability
-----------------------------------------------------------------------------*/
void TBehaviorNetwork::InsertSorted(CObList& list, TCompetence* module)
{
	POSITION			oldPos;
	POSITION			pos = list.GetHeadPosition();
	TCompetence*	compare;
	double		value = module->GetActivationAndExecutability();

	while(pos != NULL)
	{
		oldPos = pos;
		compare = (TCompetence*) list.GetNext(pos);
		if (value >= compare->GetActivationAndExecutability())
		{
			list.InsertBefore(oldPos, module);
			return;
		}
	}

	// it is the lowest object so far
	list.AddTail(module);
}


/*-----------------------------------------------------------------------------
Function: void TBehaviorNetwork::ReduceThreshold()
Date    : 19.11.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: reduces the threshold for action selection, so that a module may be executed
-----------------------------------------------------------------------------*/
void TBehaviorNetwork::ReduceThreshold()
{
	m_dThreshold -= m_pNetworkParams->GetThetaReduction();
}

/*-----------------------------------------------------------------------------
Function: void TBehaviorNetwork::UpdatePerception()
Date    : 29.01.98
-------------------------------------------------------------------------------
Parameter: true, if the new perception is from network, false, if it is replayed
Returnval: (void)
Description: calls the update function to update all propositions
-----------------------------------------------------------------------------*/
void TBehaviorNetwork::UpdatePerception(BOOL bFromNetwork)
{ 
	// update perceptions
	if (!m_bBlocksworld)
		m_pPercept->CalculateAllActivations();
	
	// if the info is replayed, refresh the network view
	if (!bFromNetwork)
		m_pNetworkView->RedrawView();
}

/*-----------------------------------------------------------------------------
Function: void TBehaviorNetwork::SetNormCoordinates()
Date    : 04.02.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: sets the origin coordinates of all modules
-----------------------------------------------------------------------------*/
void TBehaviorNetwork::SetNormCoordinates()
{
	int	i;
	int	sx = m_pNetworkView->GetScreenParams()->GetStretchX();
	int	sy = m_pNetworkView->GetScreenParams()->GetStretchY();
//	int	iMaxX = m_pNetworkView->GetScreenParams()->GetSize().cx - 30;
	int	iMaxX = 800;
	int	x;
	int	noOfColumns = 0;
	TCompetence*	pCompetence;

	// set the coordinates
	POSITION pos = m_pCompetenceList->GetHeadPosition();
	i = 0;
	while (pos != NULL)
	{
		x = (i + 1) * sx * (int) MODULE_SIZE;
		if (x > iMaxX)
		{
			// start a new column
			noOfColumns++;
			i = 0;
			x = (i + 1) * sx * (int) MODULE_SIZE;
		}
		pCompetence = (TCompetence*) m_pCompetenceList->GetNext(pos);
		pCompetence->SetNormCoordinates(CPoint(x, 70 + sy + noOfColumns * 20));
		i++;
	}

	// perception
	m_pPercept->SetAllNormCoordinates(CPoint(0, 140), CSize(sx, 2 * sy), x);

	// goals
	pos = m_pGoalList->GetHeadPosition();
	i = 0;
	while (pos != NULL)
	{
		((TGoal*) m_pGoalList->GetNext(pos))->SetNormCoordinates(
									CPoint((i + 1) * sx * (int) MODULE_SIZE, 10));
		i++;
	}
}

/*-----------------------------------------------------------------------------
Function: void TBehaviorNetwork::SetCoordinates(TScreenParamsNetwork* pScreen)
Date    : 04.02.98
-------------------------------------------------------------------------------
Parameter: pointer to the scaling and translation and so on
Returnval: (void)
Description: sets the new coordinates of all modules
-----------------------------------------------------------------------------*/
void TBehaviorNetwork::SetCoordinates(TScreenParamsNetwork* pScreen)
{
	// set the coordinates
	POSITION pos = m_pCompetenceList->GetHeadPosition();
	while (pos != NULL)
		((TCompetence*) m_pCompetenceList->GetNext(pos))->SetCoordinates(pScreen);

	// perception
	m_pPercept->SetAllCoordinates(pScreen);

	// goals
	pos = m_pGoalList->GetHeadPosition();
	while (pos != NULL)
		((TGoal*) m_pGoalList->GetNext(pos))->SetCoordinates(pScreen);
}

/*-----------------------------------------------------------------------------
Function: void TBehaviorNetwork::DrawNetwork(CDC* pDC, TScreenParamsNetwork* pScreen)
Date    : 02.02.98
-------------------------------------------------------------------------------
Parameter: pointer to the device context
			  parameters defining the screen (translation, scaling, ..)
Returnval: (void)
Description: draws the competence network or the action history
-----------------------------------------------------------------------------*/
void TBehaviorNetwork::DrawNetwork(CDC* pDC, TScreenParamsNetwork* pScreen)
{
	if (pScreen->GetDrawActionHistory())
		DrawActionHistory(pDC, pScreen);
	else
		DrawNetworkStructure(pDC, pScreen);
}

/*-----------------------------------------------------------------------------
Function: void TBehaviorNetwork::DrawNetworkStructure(CDC* pDC, TScreenParamsNetwork* pScreen)
Date    : 02.02.98
-------------------------------------------------------------------------------
Parameter: pointer to the device context
			  parameters defining the screen (translation, scaling, ..)
Returnval: (void)
Description: draws the competence network
-----------------------------------------------------------------------------*/
void TBehaviorNetwork::DrawNetworkStructure(CDC* pDC, TScreenParamsNetwork* pScreen)
{
	POSITION			pos;

	// draw the competence modules
	pos = m_pCompetenceList->GetHeadPosition();
	while (pos != NULL)
		((TCompetence*) m_pCompetenceList->GetNext(pos))->DrawModule(pDC, pScreen);

	// draw the perceptions + activation
	m_pPercept->DrawAll(pDC, pScreen);

	// draw the goals + importances
	pos = m_pGoalList->GetHeadPosition();
	while (pos != NULL)
		((TGoal*) m_pCompetenceList->GetNext(pos))->DrawModule(pDC, pScreen);
}

/*-----------------------------------------------------------------------------
Function: void TBehaviorNetwork::DrawActionHistory(CDC* pDC, TScreenParamsNetwork* pScreen)
Date    : 02.02.98
-------------------------------------------------------------------------------
Parameter: pointer to the device context
			  parameters defining the screen (translation, scaling, ..)
Returnval: (void)
Description: draws the competence network
-----------------------------------------------------------------------------*/
void TBehaviorNetwork::DrawActionHistory(CDC* pDC, TScreenParamsNetwork* pScreen)
{
	COLORREF cCol;
	CBrush	brush;
	CBrush	*oldBrush;
	int		i;
	int		border;
	int		action;
	int		x;
	int		y;
	int		size;
	POSITION	pos;
	// for names
	int		nPointSize;
	CFont		font;
	CFont*	oldFont;
	CRect		field;
	COLORREF	oldTextColor;
	COLORREF	oldBkColor;
	int		sizeX;
	CString	strName;

	border = m_lNoOfActions;
	if (border > MAX_CYCLES)
		border = MAX_CYCLES;

	// draw the action names
	nPointSize = (int) (30.0 * pScreen->GetScale());
	if (!font.CreatePointFont(nPointSize, _T("Times New Roman"), pDC))
		font.CreateStockObject(SYSTEM_FONT);
	oldFont = pDC->SelectObject(&font);
	oldBkColor = pDC->SetBkColor(0x00ffffff);		
	oldTextColor = pDC->SetTextColor(0x00000000);

	pos = m_pAct->GetFirstActionPosition();
	i = 0;
	while (pos != NULL)
	{
		strName = m_pAct->GetNextAction(pos)->GetName();
		// calculate the rectangle of the text
		sizeX = (int) pScreen->GetScale() * strName.GetLength() * nPointSize / 20;
		x = 5 + pScreen->GetTranslation().x;
		y = (int)(i * 6.0 * pScreen->GetScale()) + pScreen->GetTranslation().y;
		field.SetRect(x, y, x + sizeX, y + nPointSize / 5);
		
		pDC->DrawText(strName, field, DT_SINGLELINE | DT_VCENTER);
		i++;
	}

	pDC->SetTextColor(oldTextColor);
	pDC->SetBkColor(oldBkColor);
	pDC->SelectObject(oldFont);
	font.DeleteObject();

	// draw the history
	for (i = 0; i < border; i++)
	{
		action = m_aiActionExecuted[i];

		cCol = RGB(20, 20, 250);
		brush.CreateSolidBrush(cCol);
		oldBrush = pDC->SelectObject(&brush);

		// draw a square at the modules position
		size = (int) (2.0 * pScreen->GetScale());
		x = 60 + (int) (i * (0.5 * pScreen->GetScale())) + pScreen->GetTranslation().x;
		y = (int)(action * 6.0 * pScreen->GetScale()) + pScreen->GetTranslation().y;
		pDC->Rectangle(x, y, x + size, y + 2 * size);

		pDC->SelectObject(oldBrush);
		brush.DeleteObject();
	}
}

/*-----------------------------------------------------------------------------
Function: BOOL TBehaviorNetwork::ParseRules(LPCTSTR rules)
Date    : 15.01.98
-------------------------------------------------------------------------------
Parameter: string containing the rules
Returnval: TRUE, if at least one rule could be parsed
Description: tries to split the string into its rules and goals and parses them
				 if an error occures, the user is informed via messgebox
-----------------------------------------------------------------------------*/
BOOL TBehaviorNetwork::ParseRules(LPCTSTR rules)
{
	TCompetence*	pNewCompetence;
	TGoal*			pNewGoal;
	TParseString	str(rules);
	CString			temp;
	CString			message;
	CString			rule;
	BOOL				bWithinRule = FALSE;
	BOOL				bWithinGoal = FALSE;
	int				iLastPosition;
	int				iRuleStart;
	int				iGoalStart;

	// delete the existing network
	DeleteNetwork();
	// and initialize Worldstates
	m_pPercept->InitializeAll();

	m_bBlocksworld = FALSE;
	// search for ifs and endifs
	iLastPosition = str.GetPosition();
	while (str.ParseString(temp))
	{
		// if
		if (temp.CompareNoCase(STRING_IF) == 0)
		{
			if (!bWithinRule && !bWithinGoal)
			{
				// start of a rule detected
				iRuleStart = iLastPosition;
				bWithinRule = TRUE;
			}
			else
			{
				message.Format("If without endif or endgoal! \r\n Parsing Stopped.");
				AfxMessageBox(message);
				return FALSE;
			}
		}
		
		else if (temp.CompareNoCase(STRING_GOAL) == 0)
		{
			if (!bWithinGoal && !bWithinRule)
			{
				// start of a goal detected
				iGoalStart = iLastPosition;
				bWithinGoal = TRUE;
			}
			else
			{
				message.Format("Goal without endgoal or endif! \r\n Parsing Stopped.");
				AfxMessageBox(message);
				return FALSE;
			}
		}
		
		// endif
		else if (temp.CompareNoCase(STRING_END_IF) == 0)
		{
			if (bWithinRule)
			{
				// end of a rule detected
				rule = str.Mid(iRuleStart, str.GetPosition() - iRuleStart);	
				
				pNewCompetence = CreateNewCompetence();
				if (ParseCompetenceRule(pNewCompetence, rule))
				{
					// connect the new rule to the other modules, perceptions and goals
					pNewCompetence->ConnectCompetence(m_pCompetenceList, m_pGoalList);

					// add the new rule to the list of competences
					m_pCompetenceList->AddTail(pNewCompetence);
				}
				else
					// an error has been detected within the rule
					delete pNewCompetence;

				bWithinRule = FALSE;
			}
			else
			{
				message.Format("Endif without if! \r\n Parsing Stopped.");
				AfxMessageBox(message);
				return FALSE;
			}
		}
	
		// endgoal
		else if (temp.CompareNoCase(STRING_END_GOAL) == 0)
		{
			if (bWithinGoal)
			{
				// end of a rule detected
				rule = str.Mid(iGoalStart, str.GetPosition() - iGoalStart);	
				
				pNewGoal = CreateNewGoal();
				pNewGoal->SetIndex(m_pGoalList->GetCount());
				if (pNewGoal->ParseGoal(rule, m_pPercept))
					// add the new rule to the list of competences
					m_pGoalList->AddTail(pNewGoal);
				else
					// an error has been detected within the rule
					delete pNewGoal;

				bWithinGoal = FALSE;
			}
			else
			{
				message.Format("Endgoal without Goal! \r\n Parsing Stopped.");
				AfxMessageBox(message);
				return FALSE;
			}
		}
	
		iLastPosition = str.GetPosition();
	}

	if (GetNoOfRules() > 0)
	{
		SetNormCoordinates();
		SetCoordinates(m_pNetworkView->GetScreenParams());
		m_pNetworkView->RedrawView();
		return TRUE;
	}
	else
	{
		AfxMessageBox("No Rules detected !");
		return FALSE;
	}
}

/*-----------------------------------------------------------------------------
Function: TGoal* TBehaviorNetwork::CreateNewGoal()
Date    : 03.02.99
-------------------------------------------------------------------------------
Parameter: none
Returnval: pointer to the new goal
Description: virtual factory function
-----------------------------------------------------------------------------*/
TGoal* TBehaviorNetwork::CreateNewGoal()
{
	return new TGoal();
}

/*-----------------------------------------------------------------------------
Function: BOOL TBehaviorNetwork::ParseCompetenceRule(TCompetence* pCompetence, LPCTSTR rule)
Date    : 16.04.99
-------------------------------------------------------------------------------
Parameter: pointer to the competence module, that should contain the rule
			  string that contains the rule to parse
Returnval: (void)
Description: calls the modules parse function. Needed because ParseRule interface
				 differs from network to network
-----------------------------------------------------------------------------*/
BOOL TBehaviorNetwork::ParseCompetenceRule(TCompetence* pCompetence, LPCTSTR rule)
{
	return pCompetence->ParseRule(rule, m_pAct, m_pPercept);
}

/*-----------------------------------------------------------------------------
Function: CString TBehaviorNetwork::GetResults() const
Date    : 29.06.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: result string (see below)
Description: returns statistical results of the action selection as a string
				 thats components are separated by comma
				 if making changes to this function, you probably have to change
				 GetResultsInitial as well
-----------------------------------------------------------------------------*/
CString TBehaviorNetwork::GetResults() const
{
//	POSITION			pos;
//	TCompetence*	pCompetence;
	CString			res;

	// get the overall results
	res.Format(" %4.2f %5d", m_dAvgNoOfExecutableModules, m_lNoOfActionChanges);
/*	res.Format(" %5d %4.2f %5d %6.2f", m_lNoOfActions, m_dAvgNoOfExecutableModules
										, m_lNoOfActionChanges);
	
	// get the results from each competence module
	pos = m_pCompetenceList->GetHeadPosition();
	while (pos != NULL)
	{
		pCompetence = (TCompetence*) m_pCompetenceList->GetNext(pos);
		res += pCompetence->GetResults();
	}
*/
	#ifdef ACTIVATION_DISTRIBUTION
	{
		CFile file;
		// write out the activation distributions
		// open file
		if(file.Open("c:\\klaus\\roboklaus\\soccermonitor\\data\\distribution.dat"
						, CFile::modeCreate | CFile::modeWrite))
		{
			CArchive	archive(&file, CArchive::store);

			// perceptions
			m_pPercept->WriteActivationDistribution(archive);

			// competences
			// goals

			archive.Close();
			file.Close();
		}
	}
	#endif
	
	return res;
}

CString TBehaviorNetwork::GetResultsInitial() const
{
//	POSITION			pos;
//	TCompetence*	pCompetence;
	CString			res;

	// get the overall results
//	res = " ActionsPerformed AvgNoOfExecutableModules ActionChanges AvgDistanceToGoalWhenRelaxing";
	res = " AvgNoOfExecutableModules ActionChanges";
		
	// get the results from each competence module
/*	pos = m_pCompetenceList->GetHeadPosition();
	while (pos != NULL)
	{
		pCompetence = (TCompetence*) m_pCompetenceList->GetNext(pos);
		res += pCompetence->GetResultsInitial();
	}
*/	return res;
}

void TBehaviorNetwork::SetCombo(CComboBox* pCombo, BOOL bWhich)
{
	CString	name;

	if (bWhich)
	{
		// fill in the names of the perceptions
		POSITION pos = m_pPercept->GetFirstPerceptionPosition();
		while(pos != NULL)
		{
			name = m_pPercept->GetNextPerception(pos)->GetName();
			if (pCombo->FindString(-1, name) < 0)
				pCombo->AddString(name);
		}
	}
	else
	{
		// fill in the names of the actions
		POSITION pos = m_pAct->GetFirstActionPosition();
		while(pos != NULL)
		{
			name = m_pAct->GetNextAction(pos)->GetName();
			if (pCombo->FindString(-1, name) < 0)
				pCombo->AddString(name);
		}
	}
}

CString TBehaviorNetwork::GetActionPerformedToString()
{
	if (m_pLastAction != NULL)
		return m_pLastAction->GetName();
	else
		return(CString("No Action"));
}

int TBehaviorNetwork::GetNoOfRules() const 
{
	if (m_pCompetenceList != NULL)
		return m_pCompetenceList->GetCount();
	else
		return 0; 
}

// no inline functions, due to dll

void TBehaviorNetwork::ParamsDialog()
{
}

void TBehaviorNetwork::CalculateSpreading()
{
}

void TBehaviorNetwork::SetThreshold(double val) 
{
	m_dThreshold = val; 
}

TNetworkParams* TBehaviorNetwork::GetNetworkParams() 
{
	return m_pNetworkParams;
}

BOOL TBehaviorNetwork::IsBlocksworld()
{
	return m_bBlocksworld;
}

TCompetence* TBehaviorNetwork::CreateNewCompetence() 
{
	return NULL;
}

double TBehaviorNetwork::GetActivationSum() const
{
	return 0.0;
}

void TBehaviorNetwork::SetUseFuzzyValues(BOOL val)
{
	m_pPercept->SetUseFuzzyValues(val);
}

BOOL TBehaviorNetwork::GetUseFuzzyValues() const
{
	return m_pPercept->GetUseFuzzyValues();
}

void TBehaviorNetwork::SetFuzzyRound(double val)
{
	m_pPercept->SetFuzzyRound(val);
}



// for blocksworld

BOOL TBehaviorNetwork::PerformCycle(TNetworkParams* params, int iNoOfCycles)
{
	int	i = 0;
	BOOL	bSelectedAction;
	BOOL	bReachedGoal;

	// set the network parameters to the new ones
	m_pNetworkParams->SetNetworkParams(params);

	do
	{
		// repeat until an action has been selected
		bSelectedAction = FALSE;

		// update the goal relevances
		UpdatePerception(TRUE);

		do
		{
			// reduce the threshold
			// the base class member is called, because TDorer2Network stops
			// reducing at 0.00001 which is not what we want in blocksworld
			TBehaviorNetwork::ReduceThreshold();
			// and do a further activation spreading
			DoActivationSpreading();

			// do the action selection in old behavior network semantik selecting
			// the most active module
			bSelectedAction = TBehaviorNetwork::DoActionSelection();
		} while (!bSelectedAction);

		// reset theta
		m_dThreshold = m_pNetworkParams->GetTheta();

		// check if we reached the goal
		bReachedGoal = CheckGoalReached();

		i++;
		
	} while (!bReachedGoal && (i < iNoOfCycles));
	
	return bReachedGoal;
}

void TBehaviorNetwork::ResetNetwork()
{
	m_pRulesView->Parse();
}

BOOL TBehaviorNetwork::CheckGoalReached()
{
	BOOL		bGoalReached = TRUE;
	TGoal*	pGoal;

	POSITION pos = m_pGoalList->GetHeadPosition();
	while ((pos != NULL) && bGoalReached)
	{
		pGoal = (TGoal*) m_pGoalList->GetNext(pos);
		if (pGoal->GetCondition()->GetValue() > 0.5)
			bGoalReached = TRUE;
		else
			bGoalReached = FALSE;
	}
	return bGoalReached;
}

/*-----------------------------------------------------------------------------
Function: BOOL TBehaviorNetwork::ParseRulesWithVariables(LPCTSTR rules)
Date    : 15.01.98
-------------------------------------------------------------------------------
Parameter: string containing the rules
Returnval: TRUE, if at least one rule could be parsed
Description: parses a rule containing variables, used for blocksworld
-----------------------------------------------------------------------------*/
BOOL TBehaviorNetwork::ParseRulesWithVariables(LPCTSTR rules)
{
	TParseString	str(rules);
	CString			temp;
	CString			message;

	// delete the existing network
	DeleteNetwork();
	// and initialize Worldstates
	m_pPercept->InitializeAll();

	m_bBlocksworld = TRUE;
	// look for blocks headings
	while (str.ParseString(temp))
	{
		// objects
		if (temp.CompareNoCase(STRING_OBJECTS) == 0)
		{
			if (!ParseObjects(str))
			{
				message.Format("Error parsing objects! \r\n Parsing Stopped.");
				AfxMessageBox(message);
				return FALSE;
			}
		}

		// attributes
		else if (temp.CompareNoCase(STRING_ATTRIBUTES) == 0)
		{
			if (!ParseAttributes(str))
			{
				message.Format("Error parsing attributes! \r\n Parsing Stopped.");
				AfxMessageBox(message);
				return FALSE;
			}
		}
		
		// goals
		else if (temp.CompareNoCase(STRING_GOALS) == 0)
		{
			if (!ParseGoals(str))
			{
				message.Format("Error parsing goals! \r\n Parsing Stopped.");
				AfxMessageBox(message);
				return FALSE;
			}
		}

		// initial state
		else if (temp.CompareNoCase(STRING_INITIAL) == 0)
		{
			if (!ParseInitial(str))
			{
				message.Format("Error parsing initial state! \r\n Parsing Stopped.");
				AfxMessageBox(message);
				return FALSE;
			}
		}

		// if
		else if (temp.CompareNoCase(STRING_IF) == 0)
		{
			if (!ParseRule(str))
			{
				message.Format("Error parsing rule %d \r\n Parsing Stopped.", m_pCompetenceList->GetCount());
				AfxMessageBox(message);
				return FALSE;
			}
		}
	}

	if (m_pCompetenceList->GetCount() > 0)
	{
		SetNormCoordinates();
		SetCoordinates(m_pNetworkView->GetScreenParams());
		SortLinks();
		m_pNetworkView->RedrawView();
		return TRUE;
	}
	else
	{
		AfxMessageBox("No Rules detected !");
		return FALSE;
	}
}


/*-----------------------------------------------------------------------------
Function: int TBehaviorNetwork::ParseObjects(TParseString &str)
Date    : 23.10.98
-------------------------------------------------------------------------------
Parameter: reference to the string to parse. Parse position is after keyword 'objects'
Returnval: TRUE, if no error occured, FALSE otherwise
Description: parses the list of objects, (which is separated by commas). Objects
				 are the constants of the domain.
-----------------------------------------------------------------------------*/
BOOL TBehaviorNetwork::ParseObjects(TParseString &str)
{
	CString	temp;

	while (str.ParseString(temp))
	{
		if (temp.CompareNoCase(STRING_END_OBJECTS) == 0)
			return TRUE;

		m_pObjects->AddTail(temp);
	}
	return FALSE;
}

/*-----------------------------------------------------------------------------
Function: int TBehaviorNetwork::ParseAttributes(TParseString &str)
Date    : 23.10.98
-------------------------------------------------------------------------------
Parameter: reference to the string to parse. Parse position is after keyword 'attributes'
Returnval: TRUE, if no error occured, FALSE otherwise
Description: parses the list of attributes. Attributes can contain variables
				 which are completely instantiated with all objects. This grows 
				 exponentially with the number of variables in an attribute
-----------------------------------------------------------------------------*/
int TBehaviorNetwork::ParseAttributes(TParseString &str)
{
	TParseString	attrib;
	CString			name;
	int				leftPos;
	int				noOfArgs;
	CString			variable;

	while (str.ParseString(attrib))
	{
		if (attrib.CompareNoCase(STRING_END_ATTRIBUTES) == 0)
			return TRUE;

		// look for an opening bracket
		leftPos = attrib.Find('(');
		if (leftPos < 0)
			// zero ary attribute
			m_pPercept->AddPerception(attrib, NULL, NULL);

		else
		{
			// there is at least one argument
			name = attrib.Left(leftPos);

			// count all arguments, which can be variables or constants
			noOfArgs = 0;
			attrib.SetPosition(leftPos + 1);
			while (attrib.ParseString(variable, ','))
			{
				m_bHaveConst = FALSE;
				if ((variable.GetAt(0) <= 'Z') && (variable.GetAt(0) >= 'A'))
					noOfArgs++;
				else
				{
					m_bHaveConst = TRUE;
					m_strConst = variable;
				}
			}
			
			// create corresponding propositions (too lazy to do for any number of variables
			if (!CreatePropositions(noOfArgs, name))
				return FALSE;
		}
	}
	return FALSE;
}

/*-----------------------------------------------------------------------------
Function: BOOL TBehaviorNetwork::ParseGoals(TParseString &str)
Date    : 23.10.98
-------------------------------------------------------------------------------
Parameter: reference to the string to parse. Parse position is after keyword 'goals'
Returnval: TRUE, if no error occured, FALSE otherwise
Description: parses the list of goals 
-----------------------------------------------------------------------------*/
BOOL TBehaviorNetwork::ParseGoals(TParseString &str)
{
	AfxMessageBox("Use Dorer2 type network for goals with relevance condition !");
	return FALSE;
}

/*-----------------------------------------------------------------------------
Function: BOOL TBehaviorNetwork::ParseInitial(TParseString &str)
Date    : 23.10.98
-------------------------------------------------------------------------------
Parameter: reference to the string to parse. Parse position is after keyword 'initial'
Returnval: TRUE, if no error occured, FALSE otherwise
Description: parses the list of initially true propositions
-----------------------------------------------------------------------------*/
BOOL TBehaviorNetwork::ParseInitial(TParseString &str)
{
	CString			temp;
	CString			message;
	TWorldState*	pState;

	while (str.ParseString(temp))
	{
		if (temp.CompareNoCase(STRING_END_INITIAL) == 0)
			return TRUE;

		// lookup the proposition
		pState = m_pPercept->GetPerception(temp);
		if (pState == NULL)
		{
			message.Format("Proposition '%s' not existing! \r\n Rule ignored.", temp);
			AfxMessageBox(message);
			return FALSE;
		}
		
		// set the activation to true
		pState->SetActivation(1.0);
	}
	return FALSE;
}

/*-----------------------------------------------------------------------------
Function: BOOL TBehaviorNetwork::ParseRule(TParseString &str)
Date    : 23.10.98
-------------------------------------------------------------------------------
Parameter: reference to the string to parse. Parse position is after keyword 'if'
Returnval: TRUE, if no error occured, FALSE otherwise
Description: parses the list of goals 
-----------------------------------------------------------------------------*/
BOOL TBehaviorNetwork::ParseRule(TParseString &str)
{
	CString			temp;
	CString			variable;
	CString			object1;
	BOOL				bEffects = FALSE;
	BOOL				bReady = FALSE;
	BOOL				bNegated;
	int				leftPos;
	TParseString	action;


	m_strlConditions.RemoveAll();
	m_strlEffects.RemoveAll();
	bNegated = FALSE;
	while (!bReady && str.ParseString(temp))
	{
		if (temp.CompareNoCase(STRING_END_IF) == 0)
			bReady = TRUE;

		else if (temp.CompareNoCase(STRING_THEN) == 0)
			str.ParseString(m_strAction);

		else if (temp.CompareNoCase(STRING_EFFECT) == 0)
			bEffects = TRUE;

		else if (temp.CompareNoCase(STRING_NOT) == 0)
			bNegated = TRUE;

		else if (temp.CompareNoCase(STRING_AND) != 0)
		{
			if (bNegated)
				// dirty :-) not is encoded as 0 at the end of the name
				temp += '0';
			else	
				temp += '1';

			if (bEffects)
				m_strlEffects.AddTail(temp);
			else
				m_strlConditions.AddTail(temp);

			bNegated = FALSE;
		}
	}

	if (bReady)
	{
		//look at the name, which indicates the arity of the operator
		action = (TParseString) m_strAction;
		leftPos = action.Find('(');
		if (leftPos >= 0)
		{
			// count all arguments, which must be variables
			action.SetPosition(leftPos + 1);
			m_strlVariables.RemoveAll();
			while (action.ParseString(variable, ','))
				m_strlVariables.AddTail(variable);
			
			// action is not negated (for compatibility)
			m_strAction += '1';

			// create corresponding propositions (too lazy to do for any number of variables
			if (!CreateCompetences(m_strlVariables.GetCount()))
				return FALSE;

		}

		return TRUE;
	}
	else
		return FALSE;
}

/*-----------------------------------------------------------------------------
Function: void TBehaviorNetwork::CreatePropositions(int depth, LPCTSTR name)
Date    : 02.11.98
-------------------------------------------------------------------------------
Parameter: depth of the loop, equal to the number of arguments (must be > 0)
			  name of the proposition, that is recursively generated
Returnval: (void)
Description: generates a full instantiation of the proposition
-----------------------------------------------------------------------------*/
BOOL TBehaviorNetwork::CreatePropositions(int depth, LPCTSTR name)
{
	POSITION	pos1 = m_pObjects->GetHeadPosition();
	CString	newName;
	CString	object;
	int		where;

	depth--;
	while (pos1 != NULL)
	{
		// look if object is not allready in the string (e.g. no on-a-a)
		object = m_pObjects->GetNext(pos1);
		newName = name;
		if (((where = newName.Find('-')) < 0) || 
			 (newName.Right(newName.GetLength() - where + 1).Find(object) == -1))
		{
			// it is a new object instantiation
			newName += "-" + object;
			if (depth > 0)
			{
				// do recursion
				if (!CreatePropositions(depth, newName))
					return FALSE;
			}
			else
			{
				// we have as many loops as arguments
				if (m_bHaveConst)
					newName += "-" + m_strConst;
				m_pPercept->AddPerception(newName, NULL, NULL);
			}
		}
	}
	return TRUE;
}

/*-----------------------------------------------------------------------------
Function: void TBehaviorNetwork::CreateCompetences(int depth, LPCTSTR name)
Date    : 02.11.98
-------------------------------------------------------------------------------
Parameter: depth of the loop, equal to the number of arguments (must be > 0)
			  name of the action, that is recursively generated
Returnval: (void)
Description: generates a full instantiation of the competence modules
-----------------------------------------------------------------------------*/
BOOL TBehaviorNetwork::CreateCompetences(int depth)
{
	POSITION		pos1 = m_pObjects->GetHeadPosition();
	TCompetence	*pNewCompetence;
	CString		newName;

	depth--;
	while (pos1 != NULL)
	{
		newName = m_pObjects->GetNext(pos1);
		if (m_strlObjects.Find(newName) == NULL)
		{
			m_strlObjects.AddTail(newName);
			if (depth > 0)
				// do recursion
				CreateCompetences(depth);
			else
			{
				// rule is parsed, so create new modules for each instantiation of the rule
				pNewCompetence = CreateNewCompetence();
				if (!pNewCompetence->Instantiate(m_strlObjects, m_strlVariables, m_strlConditions
															, m_strAction, m_strlEffects, m_pPercept))
				{
					delete pNewCompetence;
					return FALSE;
				}

				// connect the new rule to the other modules, perceptions and goals
				pNewCompetence->ConnectCompetence(m_pCompetenceList, m_pGoalList);

				// set the correct maximal activity of a module
//				pNewCompetence->SetMaxActivation(MAX_ACTIVATION);

				// add the new rule to the list of competences
				m_pCompetenceList->AddTail(pNewCompetence);
			}
			m_strlObjects.RemoveTail();
		}
	}
	
	return TRUE;
}

/*-----------------------------------------------------------------------------
Function: void TBehaviorNetwork::SortLinks()
Date    : 29.10.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: calls each competence modules sort routine, to sort the links, so
				 that links to the same module are next to each other for drawing
-----------------------------------------------------------------------------*/
void TBehaviorNetwork::SortLinks()
{
	POSITION	pos = m_pCompetenceList->GetHeadPosition();

	while (pos != NULL)
		((TCompetence*) m_pCompetenceList->GetNext(pos))->SortLinks();
}

/*-----------------------------------------------------------------------------
Function: void TBehaviorNetwork::AddPerception(LPCTSTR name, CObject* pObj
												 , PCallBackPerception pFn)
Date    : 22.02.99
-------------------------------------------------------------------------------
Parameter: name of the perception
			  pointer to the object that contains the callback function
			  pointer to the callback function that calculates the truth value of
			  the perception
Returnval: (void)
Description: adds a perception and the callback to it to the perception factory TPercept
-----------------------------------------------------------------------------*/
void TBehaviorNetwork::AddPerception(LPCTSTR name, CObject* pObj
												 , PCallBackPerception pFn)
{
	m_pPercept->AddPerception(name, pObj, pFn);
}

/*-----------------------------------------------------------------------------
Function: void TBehaviorNetwork::AddAction(LPCTSTR name, CObject* pObj, PCallBackAction pFn)
Date    : 22.02.99
-------------------------------------------------------------------------------
Parameter: name of the perception
			  pointer to the object that contains the callback function
			  pointer to the callback function that executes the action
Returnval: (void)
Description: adds an action and the callback to it to the action factory TAct
-----------------------------------------------------------------------------*/
void TBehaviorNetwork::AddAction(LPCTSTR name, CObject* pObj, PCallBackAction pFn)
{
	m_pAct->AddAction(name, pObj, pFn);
}

/*-----------------------------------------------------------------------------
Function: TEnergySpreader* TBehaviorNetwork::GetModuleFromCoordinates(CPoint point, TScreenParamsNetwork* pScreen) 
Date    : 13.05.98
-------------------------------------------------------------------------------
Parameter: point, where the mouse clicked
			  screen coordinates (scaling, translation)
Returnval: (void)
Description: calculates the module that has been clicked
-----------------------------------------------------------------------------*/
TEnergySpreader* TBehaviorNetwork::GetModuleFromCoordinates(CPoint point, TScreenParamsNetwork* pScreen) 
{
	POSITION				pos;
	TEnergySpreader*	pModule;
	double				size;

	// check all competence modules (full size)
	size = MODULE_SIZE * pScreen->GetScale() * 1.5;
	pos = m_pCompetenceList->GetHeadPosition();
	while (pos != NULL)
	{
		pModule = (TEnergySpreader*) m_pCompetenceList->GetNext(pos);
		if (pModule->IsPointWithinModule(point, size))
			// we fount the module
			return pModule;
	}

	// check all goals
	pos = m_pGoalList->GetHeadPosition();
	while (pos != NULL)
	{
		pModule = (TEnergySpreader*) m_pGoalList->GetNext(pos);
		if (pModule->IsPointWithinModule(point, size))
			// we fount the module
			return pModule;
	}
		
	// check all perceptions
	size = MODULE_SIZE * pScreen->GetScale();
	pos = m_pPercept->GetFirstPerceptionPosition();
	while(pos != NULL)
	{
		pModule = m_pPercept->GetNextPerception(pos)->GetProposition()->GetModule();
		if (pModule->IsPointWithinModule(point, size))
			// we fount the module
			return pModule;
	}

	return NULL;
}
