// TMaesCompetence.cpp: implementation of the TMaesCompetence class.
//
//////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include "NetworkHeader.h"
#include "TProposition.h"
#include "TWorldState.h"
#include "TGoal.h"
#include "TNetworkParams.h"
#include "TMaesConnection.h"
#include "TMaesCompetence.h"

#include <math.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

TMaesCompetence::TMaesCompetence(TNetworkParams* params, LPCTSTR name, int index)
					:TCompetence(params, name, index)
{
	SetMaxActivation(MAX_ACTIVATION);
}

TMaesCompetence::~TMaesCompetence()
{
}

/*-----------------------------------------------------------------------------
Function: void TMaesCompetence::AddPropositionLink(TProposition* source, TProposition* dest)
Date    : 29.01.98
-------------------------------------------------------------------------------
Parameter: pointer to the proposition that is a precondition of this module
			  pointer to the destination proposition
Returnval: (void)
Description: adds an energysource that is a situation proposition
-----------------------------------------------------------------------------*/
void TMaesCompetence::AddPropositionLink(TProposition* source, TProposition* dest)
{
	// create a new connection
	TConnection* newConnection = new TMaesSituationLink(m_pNetworkParams, source, dest);
	AddEnergySource(newConnection, source, SITUATION_CONNECTION);
}

/*-----------------------------------------------------------------------------
Function: void TMaesCompetence::AddGoalLink(TProposition* source, TProposition* dest)
Date    : 29.01.98
-------------------------------------------------------------------------------
Parameter: pointer to the source proposition
			  pointer to the destination proposition
Returnval: (void)
Description: adds an energysource that is a situation proposition
-----------------------------------------------------------------------------*/
void TMaesCompetence::AddGoalLink(TProposition* source, TProposition* dest)
{
	// create a new connection
	TConnection* newConnection = new TMaesGoalLink(m_pNetworkParams, source, dest);
	AddEnergySource(newConnection, source, GOAL_CONNECTION);
}

/*-----------------------------------------------------------------------------
Function: void TMaesCompetence::AddProtectedGoalLink(TProposition* source, TProposition* dest)
Date    : 29.01.98
-------------------------------------------------------------------------------
Parameter: pointer to the source proposition
			  pointer to the destination proposition
Returnval: (void)
Description: adds an energysource that is a situation proposition
-----------------------------------------------------------------------------*/
void TMaesCompetence::AddProtectedGoalLink(TProposition* source, TProposition* dest)
{
	// create a new connection
	TConnection* newConnection = new TMaesProtectedGoalLink(m_pNetworkParams, source, dest);
	AddEnergySource(newConnection, source, PROTECTED_GOAL_CONNECTION);
}

/*-----------------------------------------------------------------------------
Function: void TMaesCompetence::AddSuccessor(TProposition* source, TProposition* dest)
Date    : 29.01.98
-------------------------------------------------------------------------------
Parameter: pointer to the source proposition
			  pointer to the destination proposition
Returnval: (void)
Description: adds an energysource that is a situation proposition
-----------------------------------------------------------------------------*/
void TMaesCompetence::AddSuccessor(TProposition* source, TProposition* dest)
{
	// create a new connection
	TConnection* newConnection = new TMaesSuccessorLink(m_pNetworkParams, source, dest);
	AddEnergySource(newConnection, source, SUCCESSOR_CONNECTION);
}

/*-----------------------------------------------------------------------------
Function: void TMaesCompetence::AddPredecessor(TProposition* source, TProposition* dest)
Date    : 29.01.98
-------------------------------------------------------------------------------
Parameter: pointer to the source proposition
			  pointer to the destination proposition
Returnval: (void)
Description: adds an energysource that is a situation proposition
-----------------------------------------------------------------------------*/
void TMaesCompetence::AddPredecessor(TProposition* source, TProposition* dest)
{
	// create a new connection
	TConnection* newConnection = new TMaesPredecessorLink(m_pNetworkParams, source, dest);
	AddEnergySource(newConnection, source, PREDECESSOR_CONNECTION);
}

/*-----------------------------------------------------------------------------
Function: void TMaesCompetence::AddConflictor(TProposition* source, TProposition* dest)
Date    : 29.01.98
-------------------------------------------------------------------------------
Parameter: pointer to the proposition (effect) of the passed module
			  pointer to the proposition (precondition) of this module
Returnval: (void)
Description: adds an energysource of which this module is a conflictor
-----------------------------------------------------------------------------*/
void TMaesCompetence::AddConflictor(TProposition* source, TProposition* dest)
{
	TMaesConflictorLink*		newConnection;
	TMaesCompetence*			pSource;
	CObList*						conflictors;
	POSITION						pos;
	TMaesConflictorLink*		pCon;

	// add the module
	newConnection = new TMaesConflictorLink(m_pNetworkParams, source, dest);
	AddEnergySource(newConnection, source, CONFLICTOR_CONNECTION);
	
	// tell the module, that this is a new conflictor of it
	pSource = (TMaesCompetence*) source->GetModule();
	
	// check for mutual conflictors
	conflictors = pSource->GetConflictors();
	pos = conflictors->GetHeadPosition();
	while (pos != NULL)
	{
		// check each conflictor connection of source
		pCon = (TMaesConflictorLink*) conflictors->GetNext(pos);
		if (pCon->GetSourceModule() == this)
		{
			// we have mutual inhibition, so mark the links as mutual inhibited
			pCon->SetMutualConflictor();
			newConnection->SetMutualConflictor();
		}
	}
}

/*-----------------------------------------------------------------------------
Function: void TMaesCompetence::CalculateMaesActivation(int iNoOfModules, double dSumOfActivation)
Date    : 08.01.98
-------------------------------------------------------------------------------
Parameter: no of modules in the network
			  sum of activation in the whole network
Returnval: (void)
Description: first the old activation is decayed
				 calculates the activation of this module. This is done by summing
				 up the energy of each incoming connection to this module according
				 to Maes. 
-----------------------------------------------------------------------------*/
void TMaesCompetence::CalculateActivation(int iNoOfModules, double dSumOfActivation)
{
	POSITION			pos;
	TConnection*	pCon;
	int				n;
	int				m;
	int				i;
	double			activation;

	// decay of old activation
	if ((dSumOfActivation > D_ZERO_N) && (dSumOfActivation < D_ZERO_P))
		m_dExternActivation = m_dMainActivation;
	else
		m_dExternActivation = m_dMainActivation * m_pNetworkParams->GetNetworkPi()
									 * iNoOfModules / dSumOfActivation; 

	// add the energy of each incoming connection
	for (i = 0; i < SUCCESSOR_CONNECTION; i++)
	{
		n = 0;
		activation = 0.0;
		pos = m_pLinks[i]->GetHeadPosition();
		while (pos != NULL)
		{
			// add the activation of this kind of link
			pCon = (TConnection*) m_pLinks[i]->GetNext(pos);
			if (i == SITUATION_CONNECTION)
				m = pCon->GetSourceProp()->GetWorldState()->GetNoOfOccurencesInPreconditions();
			else if (i == GOAL_CONNECTION)
				m = pCon->GetSourceProp()->GetWorldState()->GetNoOfOccurencesInExcitationLinks();
			else
				m = pCon->GetSourceProp()->GetWorldState()->GetNoOfOccurencesInInhibitionLinks();

			ASSERT(m > 0);
			activation += pCon->GetActivation() / m;
			n++;
		}

		// the activation is divided by the number of sources
		if (n > 1)
			activation /= (double) n;

		m_dExternActivation += activation;
	}
	
	// only in Maes
	if (m_dExternActivation < 0.0)
		m_dExternActivation = 0.0;
}

/*-----------------------------------------------------------------------------
Function: void TMaesCompetence::CalculateExecutability()
Date    : 08.01.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: calculates the executability of this module. This is done by multiplying
				 the value of each precondition. The result is powered to 1/number of 
				 preconditions
-----------------------------------------------------------------------------*/
void TMaesCompetence::CalculateExecutability()
{
	POSITION		pos;

	m_dExecutability = 1.0;
	pos = m_pPreconditions->GetHeadPosition();
	// step through each precondition
	while (pos != NULL)
	{
		// < 0.5 works, because Maes networks use binary propositions
		if (((TProposition*) m_pPreconditions->GetNext(pos))->GetValue() < 0.5)
		{
			// the first false precondition is detected
			m_dExecutability = 0.0;
			return;
		}
	}
}

/*-----------------------------------------------------------------------------
Function: void TMaesCompetence::CalculateSpreading()
Date    : 28.01.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: calculates the new activation that is gained by activation spreading
				 among the modules
-----------------------------------------------------------------------------*/
void TMaesCompetence::CalculateSpreading()
{
	int				i;
	POSITION			pos;
	TConnection*	pCon;
	int				n;
	int				m;
	double			activation;

	m_dSpreadingActivation = 0.0;
	// add the energy of each incoming connection

	// successors and predecessors
	for (i = SUCCESSOR_CONNECTION; i <= CONFLICTOR_CONNECTION; i++)
	{
		n = 0;
		activation = 0.0;
		pos = m_pLinks[i]->GetHeadPosition();
		while (pos != NULL)
		{
			pCon = (TConnection*) m_pLinks[i]->GetNext(pos);
			if (i == PREDECESSOR_CONNECTION)
				m = pCon->GetSourceProp()->GetWorldState()->GetNoOfOccurencesInPreconditions();
			else if (i == SUCCESSOR_CONNECTION)
				m = pCon->GetSourceProp()->GetWorldState()->GetNoOfOccurencesInExcitationLinks();
			else
				m = pCon->GetSourceProp()->GetWorldState()->GetNoOfOccurencesInInhibitionLinks();

			ASSERT(m > 0);
			// the activation is divided by the number of modules that contain this 
			// precondition or effect
			activation += pCon->GetActivation() / m;
			n++;
		}
		// the activation is divided by the number of sources
		if (n > 1)
			activation /= (double) n;
		m_dSpreadingActivation += activation;
	}

	m_dSpreadingActivation += m_dExternActivation;
	if (m_dSpreadingActivation < 0.0)
		m_dSpreadingActivation = 0.0;
}

/*-----------------------------------------------------------------------------
Function: void TMaesCompetence::ExecuteModule(double activation)
Date    : 16.01.98
-------------------------------------------------------------------------------
Parameter: the modules activation/executability
Returnval: (void)
Description: executes the action connected with this module
-----------------------------------------------------------------------------*/
void TMaesCompetence::ExecuteModule(double activation)
{
	TCompetence::ExecuteModule(activation);
	// Maes resets the activation of the executed module
	ResetActivation();
}

/*-----------------------------------------------------------------------------
Function: double TMaesCompetence::GetActivationAndExecutability(double dActivation, double dExecutability)
Date    : 18.11.98
-------------------------------------------------------------------------------
Parameter: activation of the module
Returnval: (void)
Description: calculates the combination of activation and executability
-----------------------------------------------------------------------------*/
double TMaesCompetence::GetActivationAndExecutability()
{
	// < 0.5 works, because Maes uses binary propositions (being 0.0 or 1.0)
	if (m_dExecutability < 0.5)
		return 0.0;
	else
		return GetActivation();
}