// TViewingAgent.cpp: implementation of the TViewingAgent class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "stdbnx.h"
#include "Basis.h"
#include "TDirection.h"
#include "BasisView.h"
#include "TScreenParams.h"
#include "TPerception.h"
#include "TThisPlayer.h"
#include "TViewingAgent.h"

#include <math.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

/*-----------------------------------------------------------------------------
Function: TViewingAgent::TViewingAgent(CBasisView* pBView, TPerceptionView* pPView
									  , TNetworkView* pNView, TRuleView* pRView)
Date    : 08.03.99
-------------------------------------------------------------------------------
Parameter: pointer to the view containing network traffic
			  pointer to the view for perception visualization
			  pointer to the view for network visualization
			  pointer to the view for behavior rules
Returnval: (void)
Description: constructor
-----------------------------------------------------------------------------*/
TViewingAgent::TViewingAgent(CBasisView* pBView, TPerceptionView* pPView
									  , TNetworkView* pNView, TRuleView* pRView)
{
	m_pView = pBView;
	m_pView->SetAgent(this);
	m_pNetworkView = pNView;
	m_pRuleView = pRView;
	m_pPerception = new TPerception(this, pPView);
	m_iPosKickOffLeft = CPoint(0,0);
	m_iPosKickOffRight = CPoint(0,0);
	m_bOrdinaryGame = FALSE;
	m_bWaitForNextSee = FALSE;
	m_pBehaviorNetwork = NULL;
	m_iIndexPosHome = 2;
	m_iMayWaitTime = 0;
	m_bHaveTurned = FALSE;

}

TViewingAgent::~TViewingAgent()
{
	delete m_pBehaviorNetwork;
	delete m_pPerception;
}

CString TViewingAgent::GetSideIdentifier()
{
	if (m_pPerception->GetSide() == LEFT_TO_RIGHT)
		return "*";
	else
		return "+";
}

/*-----------------------------------------------------------------------------
Parameter: none
Returnval: the number of time steps until this player has the ball
Description: calculates the timesteps where a player should intersect the ball
-----------------------------------------------------------------------------*/
int TViewingAgent::GetIntersectionTime(TPlayer* pAgent, int power)
{
	int				intersectionTime;
	double			dActionRadius;
	double			dBallDistance;
	TVector			tFutureBall;

	// ???
	intersectionTime = 1;
	dActionRadius = pAgent->GetFutureDistanceActive(0, power, pAgent->GetAbsDir()
											, m_pPerception->GetEffort(), m_pPerception->GetDashPowerRate());
	while (intersectionTime < MAX_FUTURE_POSITIONS)
	{
		// we calculate the intersection of the future agent position and ball position
		tFutureBall = *(m_pPerception->GetFutureBallPosition(intersectionTime));
		dBallDistance = pAgent->GetDistanceTo(tFutureBall);
		// get distance the agent can dash
		dActionRadius = pAgent->GetNextFutureDistance();

		// test if the ball will be outside
		if (m_pPerception->IsOutside(tFutureBall))
		{
			if (dBallDistance - dActionRadius < 3.0)
				// we might get the ball
				return intersectionTime;
			else
				// no intersection point
				return 1000;
		}
			
		
		if (dActionRadius + 1.0 > dBallDistance)
		{
			// we have the position
			return intersectionTime;
		}

		// ball will still be too far this setp
		intersectionTime++;
	}
	
	if (dBallDistance - dActionRadius < 3.0)
		// we might get the ball
		return intersectionTime + 3;
	else
		// no intersection point
		return 1000;
}


/*-----------------------------------------------------------------------------
Function: void TViewingAgent::SetKnowledgeType(int type)
Date    : 19.11.98
-------------------------------------------------------------------------------
Parameter: type of knowledge (MAES or DORER or DORER2)
Returnval: (void)
Description: deletes the old network and creates a new network
-----------------------------------------------------------------------------*/
void TViewingAgent::SetKnowledgeType(int type)
{
	if (m_pBehaviorNetwork != NULL)
		delete m_pBehaviorNetwork;

	m_iKnowledgeType = type;
	switch(type)
	{
	case MAES_KNOWLEDGE:
		m_pBehaviorNetwork = new TMaesNetwork(m_pNetworkView, m_pRuleView);
		break;

	case DORER_KNOWLEDGE:
		m_pBehaviorNetwork = new TDorerNetwork(m_pNetworkView, m_pRuleView);
		break;

	case DORER2_KNOWLEDGE:
		m_pBehaviorNetwork = new TDorer2Network(m_pNetworkView, m_pRuleView);
		break;

	default:
	#ifdef _DEBUG
		TRACE("Knowledge type not existing\r\n");
	#endif
		ASSERT(FALSE);		
	}

	// add the perception functions
	CreatePerceptions();

	// add resource functions
	CreateResources();
}

/*-----------------------------------------------------------------------------
Function: void TViewingAgent::SetViewParams(int angle, int quality)
Date    : 20.01.98
-------------------------------------------------------------------------------
Parameter: view angle (one of 180, 90 or 45)
			  view quality (0 for low, 1 for high)
Returnval: void
Description: sets the internal representation of the viewing angle and quality
-----------------------------------------------------------------------------*/
void TViewingAgent::OnChangedView(int angle, int quality)
{ 
	m_pPerception->SetViewParams(angle, quality); 
}

/*-----------------------------------------------------------------------------
Function: void TViewingAgent::SwitchViewQuality()
Date    : 19.12.97
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: switches the view quality from high to low
-----------------------------------------------------------------------------*/
void TViewingAgent::SwitchViewQuality()
{
	int iQuality = m_pPerception->SwitchViewQuality();
	int iAngle = m_pPerception->GetViewAngle();
	ChangeView(iAngle, iQuality);
}

/*-----------------------------------------------------------------------------
Function: void TViewingAgent::SwitchViewAngle()
Date    : 19.12.97
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: switches from a view of 180 to 90 to 45 degrees
-----------------------------------------------------------------------------*/
void TViewingAgent::SwitchViewAngle()
{
	int iAngle = m_pPerception->SwitchViewAngle();
	int iQuality = m_pPerception->GetViewQuality();
	ChangeView(iAngle, iQuality);
}

/*-----------------------------------------------------------------------------
Function: void CBaseAgent::SetNumber(int num)
Date    : 17.03.98
-------------------------------------------------------------------------------
Parameter: number of the player
			  side the player is playing (LEFT_TO_RIGHT or RIGHT_TO_LEFT)
Returnval: (void)
Description: sets the number of the player and its position before playon
				 it is assumed, that team has been set before
-----------------------------------------------------------------------------*/
void TViewingAgent::SetNumber(int number, int side)
{ 
	int	posLX[PLAYERS_PER_TEAM] = POS_KICKOFF_LEFT_X;
	int	posLY[PLAYERS_PER_TEAM] = POS_KICKOFF_LEFT_Y;
	int	posRX[PLAYERS_PER_TEAM] = POS_KICKOFF_RIGHT_X;
	int	posRY[PLAYERS_PER_TEAM] = POS_KICKOFF_RIGHT_Y;
	int	homeX[PLAYERS_PER_TEAM] = POS_HOME_X;
	int	homeY[PLAYERS_PER_TEAM] = POS_HOME_Y;

	if (side == LEFT_TO_RIGHT)
	{
		// set the home position left to right player
		m_iPosHome.x = homeX[number - 1];
		m_iPosHome.y = homeY[number - 1];
	}
	else
	{
		// set the home position for a right to left player
		m_iPosHome.x = -homeX[number - 1];
		m_iPosHome.y = -homeY[number - 1];
	}

	// set the positions of the player
	if (side == LEFT_TO_RIGHT)
	{
		// it is a player of the left team
		m_iPosKickOffLeft.x = posLX[number - 1];
		m_iPosKickOffLeft.y = posLY[number - 1];
		m_iPosKickOffRight.x = posRX[number - 1];
		m_iPosKickOffRight.y = posRY[number - 1];
	}
	else
	{
		// it is a player of the right team, so mirror x coordinates
		m_iPosKickOffLeft.x = posRX[number - 1];
		m_iPosKickOffLeft.y = posRY[number - 1];
		m_iPosKickOffRight.x = posLX[number - 1];
		m_iPosKickOffRight.y = posLY[number - 1];
	}

	CBaseAgent::SetNumber(number, side);
}

/*-----------------------------------------------------------------------------
Function: void TViewingAgent::Turn(int iMoment, BOOL bQueue)
Date    : 07.09.98
-------------------------------------------------------------------------------
Parameter: relative angle to turn ranging from -180 to 180 where -90 is 90 degrees to the left
			  flag indicating, whether the action should be performed instantaniously
			  (FALSE, default) or queued and executed next timestep
Returnval: (void)
Description: takes into accoun the speed of the agent. the agent actually turns
				 less then iMoment, if it is fast. so this function increases turn moment
-----------------------------------------------------------------------------*/
void TViewingAgent::Turn(int iMoment, BOOL bQueue)
{
	double		speed = m_pPerception->GetSpeed();

	// increase turn
	double	dTurn = iMoment;
	dTurn *= 1.0 + speed * INERTIA_MOMENT;
	iMoment = (int)dTurn;
	if (iMoment > 180)
		iMoment = 180;
	else if (iMoment < -180)
		iMoment = -180;

	CBaseAgent::Turn(iMoment, bQueue);
}

/*-----------------------------------------------------------------------------
Function: void TViewingAgent::Kick(int iPower, int iDirection, BOOL bQueue = FALSE)
Date    : 04.11.99
-------------------------------------------------------------------------------
Parameter: power with which to kick
			  relative direction to the body
			  flag if action should be queued
Returnval: (void)
Description: introduced to give noise into the kick for dis
-----------------------------------------------------------------------------*/
void TViewingAgent::Kick(int iPower, int iDirection, BOOL bQueue)
{
	TDirection dir;

	#ifdef ACTION_NOISE_ON
		double	dNoise = m_pBehaviorNetwork->GetNetworkParams()->GetPhi();
		double	dRand = (1.0 - (double) rand() / ((double) RAND_MAX / 2.0));

		#ifdef PERCENTUAL_NOISE_ON
			// prozentualer Fehler
			iDirection += (int) ((double) iDirection * dRand * dNoise);
			iPower += iPower * (int)(dRand * dNoise);
		#else
			// absoluter Fehler
			iDirection += (int) (180.0 * dRand * dNoise);
			iPower += (int)(100.0 * dRand * dNoise);
		#endif

		if (iPower > 100)
			iPower = 100;
		if (iPower < 0)
			iPower = 0;
	#endif

	dir = iDirection;
	CBaseAgent::Kick(iPower, dir, bQueue);
}

/*-----------------------------------------------------------------------------
Function: void TViewingAgent::TurnBody(int iMoment, BOOL bQueue)
Date    : 18.06.99
-------------------------------------------------------------------------------
Parameter: relative angle to turn the body of the agent 
			  flag indicating, whether the action should be performed instantaniously
			  (FALSE, default) or queued and executed next timestep
Returnval: (void)
Description: takes into accoun the speed of the agent. the agent actually turns
				 less then iMoment, if it is fast. so this function increases turn moment
-----------------------------------------------------------------------------*/
void TViewingAgent::TurnBody(int iMoment, BOOL bQueue)
{
	TDirection	iTurn(iMoment + m_pPerception->GetHeadAngle());
	Turn(iTurn, bQueue);
}

/*-----------------------------------------------------------------------------
Function: BOOL TViewingAgent::OnConnected()
Date    : 11.05.99
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: called after the client connected to the server. Starts the timeout
				 timer for automatical shutdown, if the server does not send messages
-----------------------------------------------------------------------------*/
void TViewingAgent::OnConnected()
{
	CBaseAgent::OnConnected();
}

/*-----------------------------------------------------------------------------
Function: void CBaseAgent::OnDisconnect()
Date    : 11.05.99
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: sends the bye message to the server
-----------------------------------------------------------------------------*/
void TViewingAgent::OnDisconnect()
{
	if ((m_pView != NULL) && ::IsWindow(m_pView->m_hWnd))
		m_pView->StopTimeoutTimer();
	CBaseAgent::OnDisconnect();
}

/*-----------------------------------------------------------------------------
Function: void TViewingAgent::OnReceivedMessage(LPCTSTR message)
Date    : 16.12.97
-------------------------------------------------------------------------------
Parameter: message that has been received from the server
Returnval: void
Description: calls the TPerception ParseMessage method to parse the message
				 calls the views display message function
-----------------------------------------------------------------------------*/
void TViewingAgent::OnReceivedMessage(LPCTSTR message, BOOL bFromNetwork)
{
	CString msg = "";

	// parse the message
	m_pPerception->NewMessage(message);

	// display the message in the message view
	if (!bFromNetwork)
		m_pView->Erase();
	m_pView->DisplayDebug(msg + message, KATEGORY_PERCEPTION, DEBUG_PERCEPTION);

	// start the timeout timer if init message
	m_pView->SetReceivedData(TRUE);
	if ((message[1] == 'i') && bFromNetwork)
		m_pView->StartTimeoutTimer();

}

/*-----------------------------------------------------------------------------
Function: void TViewingAgent::OnSentMessage(LPCTSTR message)
Date    : 17.12.97
-------------------------------------------------------------------------------
Parameter: message that has been sent to the server
Returnval: (void)
Description: displays the message in the message view
-----------------------------------------------------------------------------*/
void TViewingAgent::OnSentMessage(LPCTSTR message)
{
	if ((m_pView != NULL) && ::IsWindow(m_pView->m_hWnd))
		m_pView->DisplayDebug(message, KATEGORY_ACTION, DEBUG_ACTION);
}


/*-----------------------------------------------------------------------------
Function: void TViewingAgent::OnMoved(int x, int y)
Date    : 26.03.98
-------------------------------------------------------------------------------
Parameter: position the player tried to move to
Returnval: (void)
Description: called after the agent performed a move command
-----------------------------------------------------------------------------*/
void TViewingAgent::OnMoved(int x, int y)
{
	// update the position of this player
	TPlayer*	pPlayer;

	pPlayer = m_pPerception->GetThisPlayer();
	if (GetSide() == RIGHT_TO_LEFT)
	{
		pPlayer->SetAbsX(-x);
		pPlayer->SetAbsY(-y);
	}
	else
	{
		pPlayer->SetAbsX(x);
		pPlayer->SetAbsY(y);
	}
	// tell the player, that all the other info is out of date
	m_pPerception->InvalidateAll();
}

/*-----------------------------------------------------------------------------
Function: void TViewingAgent::OnDashed(int power)
Date    : 27.08.98
-------------------------------------------------------------------------------
Parameter: power of the dash (-30 - 100)
Returnval: (void)
Description: called after a dash command. updates the stamina of the player
-----------------------------------------------------------------------------*/
void TViewingAgent::OnDashed(int power)
{
	m_pPerception->OnDashed(power);
}

/*-----------------------------------------------------------------------------
Function: void TViewingAgent::OnKicked(int power, int direction)
Date    : 03.09.98
-------------------------------------------------------------------------------
Parameter: power and relative direction (to the body) of the kick command
Returnval: (void)
Description: adjusts the ball's speed
-----------------------------------------------------------------------------*/
void TViewingAgent::OnKicked(int power, int direction)
{
	m_pPerception->OnKicked(power, direction);	
}

/*-----------------------------------------------------------------------------
Function: void TViewingAgent::OnTurned(int dir)
Date    : 26.03.98
-------------------------------------------------------------------------------
Parameter: moment with which the player turned. this may be more than the player turned
			  actually, especially if the player has got high speed
Returnval: (void)
Description: called after the agent performed a turn command
-----------------------------------------------------------------------------*/
void TViewingAgent::OnTurned(int dir)
{
	double	speed = m_pPerception->GetSpeed();

	// turn has got some inertia, so actual turn is less than dir
	double	dTurn = dir;
	dTurn /= 1.0 + speed * INERTIA_MOMENT;
	dir = (int)dTurn;

	// update the direction of this player
	m_pPerception->OnTurned(dir);
	m_bHaveTurned = TRUE;
}

/*-----------------------------------------------------------------------------
Function: void TViewingAgent::OnCaught(int dir)
Date    : 29.07.98
-------------------------------------------------------------------------------
Parameter: direction of the catch command
Returnval: (void)
Description: called, when the agent executed a catch command
-----------------------------------------------------------------------------*/
void TViewingAgent::OnCaught(int dir)
{
	m_pPerception->SetLastCatchTime();
}

/*-----------------------------------------------------------------------------
Function: void TViewingAgent::OnTurnedNeck(int dir)
Date    : 26.03.98
-------------------------------------------------------------------------------
Parameter: angle with which the player turned the neck. 
Returnval: (void)
Description: called after the agent performed a turn neck command
-----------------------------------------------------------------------------*/
void TViewingAgent::OnTurnedNeck(int dir)
{
	m_pPerception->OnTurnedNeck(dir);
}


/*-----------------------------------------------------------------------------
Description: returns the time when the agent may say something next
-----------------------------------------------------------------------------*/
int TViewingAgent::GetNextSayTime() const
{
	int time = m_pPerception->GetSenseTime();
	int	next = 0;

	// old!! the players say in order 1 5 9 2 6 10 3 7 11 4 8 every 22 cycles
	// the players say in order 1 6 11 5 10 4 9 3 8 2 7 every 22 cycles
	while (((time % 2) || (((time * 3) % 11) + 1 != m_iNumber))
		&& (next < 22))
	{
		time++;
		next++;
	}

	return time;
}

/*-----------------------------------------------------------------------------
Function: functions passed to TPerception and TBehaviorNetwork
Date    : 26.03.98
-----------------------------------------------------------------------------*/
int TViewingAgent::GetSide()
{
	return m_pPerception->GetSide();
}

void TViewingAgent::SetSide(int side)
{
	m_pPerception->SetSide(side);
}

int TViewingAgent::GetStamina() const
{
	return m_pPerception->GetStamina();
}

int TViewingAgent::GetSenseTime() const
{
	return m_pPerception->GetSenseTime();
}

void TViewingAgent::ResetNetwork()
{
	m_pBehaviorNetwork->ResetNetwork();
}

TNetworkParams* TViewingAgent::GetNetworkParams()
{
	return m_pBehaviorNetwork->GetNetworkParams();
}

int TViewingAgent::GetNoOfRules() const
{
	return m_pBehaviorNetwork->GetNoOfRules();
}

void TViewingAgent::DisplayBallPosition(TVector& pos, TVector& speed, BOOL bSee)
{
	m_pView->DisplayBallPosition(pos, speed, bSee);
}

void TViewingAgent::DisplayPlayerPosition(TVector& pos, int number, BOOL bSee)
{
	m_pView->DisplayPlayerPosition(pos, number, bSee);
}

void TViewingAgent::DisplayPosition(TVector& pos, int bodyDir, int dir, BOOL bSee)
{
	m_pView->DisplayPosition(pos, bodyDir, dir, bSee);
}

void TViewingAgent::DisplayDebug(LPCTSTR msg, int category, int level)
{
	m_pView->DisplayDebug(msg, category, level);
}

void TViewingAgent::SetNoise(double noise)
{
	m_pPerception->SetNoise(noise);
}

int	TViewingAgent::GetPlayerType()
{ 
	return m_pPerception->GetCurrentPlayerType(); 
}


/*-----------------------------------------------------------------------------
Function: double TViewingAgent::GetLinearFuzzyValue(double value, double left, double right, int up)
Date    : 13.03.98
-------------------------------------------------------------------------------
Parameter: beginning point of increase (decrease); values less than left return 0.0 (1.0)
			  end point of increase (decrease); values greater than right return 1.0 (0.0)
Returnval: fuzzy value 0.0 <= val <= 1.0
Description: calculates a fuzzy value of value as linear increase (up = TRUE default)
				 or decrease (up = FALSE) between left and right border
-----------------------------------------------------------------------------*/
double TViewingAgent::GetLinearFuzzyValue(double value, double left, double right, int up)
{
	ASSERT(left < right);

	if (up)
		if (value < left)
			return 0.0;
		else if (value > right)
			return 1.0;
		else
			return (value - left) / (right - left);
	else
		if (value < left)
			return 1.0;
		else if (value > right)
			return 0.0;
		else
			return (right - value) / (right - left);
}

int TViewingAgent::GetPassPower(double distance)
{
	return 70 + (int) (GetLinearFuzzyValue(distance, 5.0, 15.0) * 30.0);
}

	
/*-----------------------------------------------------------------------------
Function: void TViewingAgent::CreatePerceptions()
Date    : 22.02.99
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: adds the perception functions to the behavior networks function factory
-----------------------------------------------------------------------------*/
void TViewingAgent::CreatePerceptions()
{
	m_pBehaviorNetwork->AddPerception("ballAhead", this, (PCallBackPerception) GetBallAhead);
	m_pBehaviorNetwork->AddPerception("ballFarAhead", this, (PCallBackPerception) GetBallFarAhead);
	m_pBehaviorNetwork->AddPerception("ballBehind", this, (PCallBackPerception) GetBallBehind);
	m_pBehaviorNetwork->AddPerception("ballIsCatchable", this, (PCallBackPerception) GetBallIsCatchable);
	m_pBehaviorNetwork->AddPerception("ballIsFar", this, (PCallBackPerception) GetBallIsFar);
	m_pBehaviorNetwork->AddPerception("ballIsNearOtherGoal", this, (PCallBackPerception) GetBallIsNearOtherGoal);
	m_pBehaviorNetwork->AddPerception("ballIsNearOwnGoal", this, (PCallBackPerception) GetBallIsNearOwnGoal);
	m_pBehaviorNetwork->AddPerception("ballIsSlow", this, (PCallBackPerception) GetBallIsSlow);
	m_pBehaviorNetwork->AddPerception("ballIsVeryNearOwnGoal", this, (PCallBackPerception) GetBallIsVeryNearOwnGoal);
	m_pBehaviorNetwork->AddPerception("ballKickable", this, (PCallBackPerception) GetBallKickable);
	m_pBehaviorNetwork->AddPerception("ballOutside", this, (PCallBackPerception) GetBallOutside);
	m_pBehaviorNetwork->AddPerception("catchSuccess", this, (PCallBackPerception) GetCatchSuccess);
	m_pBehaviorNetwork->AddPerception("freeTeammateIsOffside", this, (PCallBackPerception) GetFreeTeammateIsOffside);
	m_pBehaviorNetwork->AddPerception("haveBall", this, (PCallBackPerception) GetHaveBall);
	m_pBehaviorNetwork->AddPerception("haveEnoughStamina", this, (PCallBackPerception) GetHaveEnoughStamina);
	m_pBehaviorNetwork->AddPerception("haveSeenBall", this, (PCallBackPerception) GetHaveSeenBall);
	m_pBehaviorNetwork->AddPerception("haveTime", this, (PCallBackPerception) GetHaveTime);
	m_pBehaviorNetwork->AddPerception("highStamina", this, (PCallBackPerception) GetHighStamina);
	m_pBehaviorNetwork->AddPerception("inOffside", this, (PCallBackPerception) GetInOffside);
	m_pBehaviorNetwork->AddPerception("inOtherHalf", this, (PCallBackPerception) GetInOtherHalf);
	m_pBehaviorNetwork->AddPerception("justCaughtBall", this, (PCallBackPerception) GetJustCaughtBall);
	m_pBehaviorNetwork->AddPerception("knowNotInOffside", this, (PCallBackPerception) GetKnowNotInOffside);
	m_pBehaviorNetwork->AddPerception("lowStamina", this, (PCallBackPerception) GetLowStamina);
	m_pBehaviorNetwork->AddPerception("maySaySomething", this, (PCallBackPerception) GetMaySaySomething);
	m_pBehaviorNetwork->AddPerception("mayWait", this, (PCallBackPerception) GetMayWait);
	m_pBehaviorNetwork->AddPerception("nearBall", this, (PCallBackPerception) GetNearBall);
	m_pBehaviorNetwork->AddPerception("nearGoal", this, (PCallBackPerception) GetNearGoal);
	m_pBehaviorNetwork->AddPerception("nearOpponent", this, (PCallBackPerception) GetNearOpponent);
	m_pBehaviorNetwork->AddPerception("nearOwnGoal", this, (PCallBackPerception) GetNearOwnGoal);
	m_pBehaviorNetwork->AddPerception("noOtherGoal", this, (PCallBackPerception) GetNoOtherGoal);
	m_pBehaviorNetwork->AddPerception("offsideTrap", this, (PCallBackPerception) GetOffsideTrap);
	m_pBehaviorNetwork->AddPerception("opponentInFront", this, (PCallBackPerception) GetOpponentInFront);
	m_pBehaviorNetwork->AddPerception("playmodeGoalkickOther", this, (PCallBackPerception) GetPlaymodeGoalkickOther);
	m_pBehaviorNetwork->AddPerception("playmodeOn", this, (PCallBackPerception) GetPlaymodeOn);
	m_pBehaviorNetwork->AddPerception("soccergoal", this, (PCallBackPerception) GetSoccergoal);
	m_pBehaviorNetwork->AddPerception("standFree", this, (PCallBackPerception) GetStandFree);
	m_pBehaviorNetwork->AddPerception("teammateBehind", this, (PCallBackPerception) GetTeammateBehind);
	m_pBehaviorNetwork->AddPerception("teammateHasBall", this, (PCallBackPerception) GetTeammateHasBall);
	m_pBehaviorNetwork->AddPerception("teammateInFront", this, (PCallBackPerception) GetTeammateInFront);
	m_pBehaviorNetwork->AddPerception("teammateIsBetweenBall", this, (PCallBackPerception) GetTeammateIsBetweenBall);
	m_pBehaviorNetwork->AddPerception("teammateIsFree", this, (PCallBackPerception) GetTeammateIsFree);
	m_pBehaviorNetwork->AddPerception("teammateIsNearerBall", this, (PCallBackPerception) GetTeammateIsNearerBall);
	m_pBehaviorNetwork->AddPerception("wayToGoalFree", this, (PCallBackPerception) GetWayToGoalFree);

	// inserted for compatibility with agent3
	m_pBehaviorNetwork->AddPerception("teammateIsNearerOtherGoal", this, (PCallBackPerception) GetTeammateIsNearerOtherGoal);
	m_pBehaviorNetwork->AddPerception("opponentIsNearerToBall", this, (PCallBackPerception) GetOpponentIsNearerToBall);
	m_pBehaviorNetwork->AddPerception("opponentIsNearerToOwnGoal", this, (PCallBackPerception) GetOpponentIsNearerToOwnGoal);

/*	// inserted for bewerbung
	m_pBehaviorNetwork->AddPerception("machtSpass", this, (PCallBackPerception) GetTrue);
	m_pBehaviorNetwork->AddPerception("istEintraeglich", this, (PCallBackPerception) GetTrue);
	m_pBehaviorNetwork->AddPerception("bietetZukunftschancen", this, (PCallBackPerception) GetTrue);
	m_pBehaviorNetwork->AddPerception("bietetFortbildung", this, (PCallBackPerception) GetTrue);
	m_pBehaviorNetwork->AddPerception("bietetSozialeLeistungen", this, (PCallBackPerception) GetTrue);
	m_pBehaviorNetwork->AddPerception("teurerStandort", this, (PCallBackPerception) GetTrue);
*/
}


/*-----------------------------------------------------------------------------
Function: Perception functions
Date    : 14.01.98
-------------------------------------------------------------------------------
Parameter: do not get parameters
Returnval: activation (truth) of this perception
Description: this are the functions that calculate the truth of perceptions
-----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
returns true, if the ball is nearer to the other goal (only x coordinate)
than this player is
-----------------------------------------------------------------------------*/
double TViewingAgent::GetBallAhead()
{
	int	x1 = (int) m_pPerception->GetThisPlayer()->GetAbsX();
	int	x2 = (int) m_pPerception->GetBall()->GetAbsX();

	if (m_pPerception->GetSide() == LEFT_TO_RIGHT)
	{
		if (/*(x2 < 47) && */(x2 > x1))
			return GetLinearFuzzyValue(x2 - x1, 0.0, 3.0);
		else
			return 0.0;
	}
	else if (/*(x2 > -47) && */(x2 < x1))
			return GetLinearFuzzyValue(x1 - x2, 0.0, 3.0);
		else
			return 0.0;
}

/*-----------------------------------------------------------------------------
returns true, if the ball is much nearer to the other goal (only x coordinate)
than this player is
-----------------------------------------------------------------------------*/
double TViewingAgent::GetBallFarAhead()
{
	int	x1 = (int) m_pPerception->GetThisPlayer()->GetAbsX();
	int	x2 = (int) m_pPerception->GetBall()->GetAbsX();

	if (m_pPerception->GetSide() == LEFT_TO_RIGHT)
	{
		if (/*(x2 < 47) &&*/ (x2 > x1))
			return GetLinearFuzzyValue(x2 - x1, 7.0, 20.0);
		else
			return 0.0;
	}
	else if (/*(x2 > -47) &&*/ (x2 < x1))
			return GetLinearFuzzyValue(x1 - x2, 7.0, 20.0);
		else
			return 0.0;
}

/*-----------------------------------------------------------------------------
returns true, if the ball is nearer to the own goal (only x coordinate)
than this player is
-----------------------------------------------------------------------------*/
double TViewingAgent::GetBallBehind()
{
	int	x1 = (int) m_pPerception->GetThisPlayer()->GetAbsX();
	int	x2 = (int) m_pPerception->GetBall()->GetAbsX();

	if (m_pPerception->GetSide() == LEFT_TO_RIGHT)
	{
//		if (x2 > -47)
			return GetLinearFuzzyValue(x1 - x2, -4.0, 0.0);
//		else
//			return 0.0;
	}
	else //if (x2 < 47)
			return GetLinearFuzzyValue(x2 - x1, -4.0, 0.0);
//		else
//			return 0.0;
}

/*-----------------------------------------------------------------------------
calculates if the ball is or will be within reach to catch (goalie)
-----------------------------------------------------------------------------*/
double TViewingAgent::GetBallIsCatchable()
{
	int				futureTime = 0;
	TVector			ballInFuture;
	TVector			toBall;
	TBall*			pBall = m_pPerception->GetBall();
	TThisPlayer*	pAgent = m_pPerception->GetThisPlayer();
	double			ballDist;
	double			ballDist2;
	CString			msg;
	double			ballX;
	double			ballY;

	// get distance of ball
	ballInFuture = *m_pPerception->GetFutureBallPosition(futureTime);

	// check if ball is in penalty area
	ballX = ballInFuture.GetX();
	ballY = ballInFuture.GetY();
	if (GetSide() == RIGHT_TO_LEFT)
	{
		ballX = -ballX;
		ballY = -ballY;
	}

	if ((ballX > -36.0) || (ballY > 16.5) || (ballY < -16.5))
		// ball is not in penalty area
		return 0.0;

	if (pBall->IsValid())
		ballDist = pBall->GetDistance();
	else
		ballDist = pAgent->GetDistanceTo(ballInFuture);

	// if it is rather far
	if ((ballDist < CATCHABLE_AREA_L) && (ballDist > CATCHABLE_AREA_L - 0.2))
	{
		// look if the ball is approaching
		ballInFuture = *m_pPerception->GetFutureBallPosition(futureTime + 1);
		ballDist2 = pAgent->GetDistanceTo(ballInFuture);
		// check if a player is close the ball
		// in that case we take a risk and catch
		if ((ballDist2 < ballDist - 0.1) && !m_pPerception->GetOpponentNearBall(2.0))
		{
			// the ball will be closer next cycle
			msg.Format("Ball catchable but will be closer and no opponent near: distance: %4.2f distance2: %4.2f"
							, ballDist, ballDist2);
			m_pView->DisplayDebug(msg, KATEGORY_PERCEPTION, DEBUG_LOWPERCEPTION);
			return 0.0;
		}
	}

	// we do not catch to farest distance because of inaccuracies
	if (ballDist > CATCHABLE_AREA_L)
		return 0.0;
	else
	{
		msg.Format("Ball catchable: distance: %4.2f distance2: %4.2f", ballDist, ballDist2);
		m_pView->DisplayDebug(msg, KATEGORY_PERCEPTION, DEBUG_LOWPERCEPTION);

		return 1.0;
	}
}

/*-----------------------------------------------------------------------------
false up to 15 m of ball distance to agent, true at 25 or greater distance
-----------------------------------------------------------------------------*/
double TViewingAgent::GetBallIsFar()
{
	if (!m_pPerception->GetBall()->IsInMind())
		return 0.0;
	else
		return GetLinearFuzzyValue(m_pPerception->GetObjectDistance(m_pPerception->GetBall()), 15.0, 25.0);
}

/*-----------------------------------------------------------------------------
perception proposition for the distance of the ball to the other goal
-----------------------------------------------------------------------------*/
double TViewingAgent::GetBallIsNearOtherGoal()
{
	return GetLinearFuzzyValue(m_pPerception->GetBall()->GetDistanceTo(
											m_pPerception->GetOtherGoal()), 15.0, 70.0, FALSE);
}

/*-----------------------------------------------------------------------------
perception proposition for the distance of the ball to the own goal
-----------------------------------------------------------------------------*/
double TViewingAgent::GetBallIsNearOwnGoal()
{
	return GetLinearFuzzyValue(m_pPerception->GetBall()->GetDistanceTo(
											m_pPerception->GetOwnGoal()), 15.0, 70.0, FALSE);
}

/*-----------------------------------------------------------------------------
perception proposition for the speed of the ball
-----------------------------------------------------------------------------*/
double TViewingAgent::GetBallIsSlow()
{
	TThisPlayer*	pAgent = m_pPerception->GetThisPlayer();
	TBall*			pBall = m_pPerception->GetBall();
	TVector			tBallSpeed = pBall->GetRelSpeed(pAgent->GetPosition(), pAgent->GetAbsDir());
	double			speed = tBallSpeed.GetLength();

	return GetLinearFuzzyValue(speed, 0.0, 0.5, FALSE);
}

/*-----------------------------------------------------------------------------
perception proposition for the (future) distance of the ball to the own goal
-----------------------------------------------------------------------------*/
double TViewingAgent::GetBallIsVeryNearOwnGoal()
{
	TVector*	futureBall = m_pPerception->GetFutureBallPosition(14);
	double	futureBallDistance = m_pPerception->GetOwnGoal()->GetDistanceTo(*futureBall);
/*	TPlayer*	opponentAtBall = m_pPerception->GetOpponentAtBall();

	// if the ball is farer, but we will be first at ball return true
	if (opponentAtBall != NULL)
	{
		double	opponentBallDistance = opponentAtBall->GetDistanceTo(*futureBall);
		if (futureBallDistance < opponentBallDistance)
			return GetLinearFuzzyValue(futureBallDistance, 16.0, 30.0, FALSE);
	}
*/
	// check if this will be inside the penalty area
	double x = futureBall->GetX();
	double y = futureBall->GetY();

	if (GetSide() == RIGHT_TO_LEFT)
	{
		x = -x;
		y = -y;
	}

	if ((x > -36.0) || (y < -16.0) || (y > 16.0))
		// outside penalty area
		return GetLinearFuzzyValue(futureBallDistance, 16.0, 30.0, FALSE);
	else
		// inside penalty area
		return 1.0;
}

/*-----------------------------------------------------------------------------
true if the ball is within the kickable_area of this agent
-----------------------------------------------------------------------------*/
double TViewingAgent::GetBallKickable()
{
	double	maxKickDistance = m_pPerception->GetPlayerSize() + BALL_SIZE + m_pPerception->GetKickableMargin();	
	double	dBallDistance = m_pPerception->GetObjectDistance(m_pPerception->GetBall());

	if ((dBallDistance > maxKickDistance -0.2) && (dBallDistance < maxKickDistance + 0.2) 
		&& m_pPerception->GetBall()->IsValid())
		dBallDistance = m_pPerception->GetBall()->GetDistance();

	if (dBallDistance > maxKickDistance)
		// ball is too far
		return 0.0;
	else 
	{
		// the ball is kickable
		return GetHaveSeenBall();
	}


/*
	TVector	ballNow;
	TVector	ballSpeed;
	TVector	intersection;
	TVector	line;
	TVector	orthogonal;
	double	dMinDist = 10000.0;
	double	maxKickDistance = PLAYER_SIZE + BALL_SIZE + KICKABLE_MARGIN;	
	double	s = 0.0;
	double	dBallAbsSpeed;

	// calculate the trajectory the ball takes
	ballNow = m_pPerception->GetBall()->GetPosition();
	ballSpeed = m_pPerception->GetBall()->GetAbsSpeed();

	// calculate closest point to ball trajectory 
	intersection = ballNow;
//	intersection = *m_pPerception->GetFutureBallPosition(1);
	dBallAbsSpeed = ballSpeed.GetLength();
	if (dBallAbsSpeed > 0.1)
	{
		// ball has significant speed
		double	px, py;		// ball coordinates
		double	lx, ly;		// direction of ball speed
		double	rx, ry;		// position of this agent
		double	ratio;		// oy/ox where vector o is orthogonal to line
		TVector	agentNextTime;

		px = ballNow.GetX();
		py = ballNow.GetY();
		lx = ballSpeed.GetX();
		ly = ballSpeed.GetY();
		agentNextTime = m_pPerception->GetThisPlayer()->GetPosition() +
						m_pPerception->GetThisPlayer()->GetAbsSpeed();
		rx = agentNextTime.GetX();
		ry = agentNextTime.GetY();
		
		if ((ly > D_ZERO_N) && (ly < D_ZERO_P))
			// the ball is approaching us
			s = (rx - px) / lx;
		else
		{
			ratio = -lx/ly;		// is orthogonal to trajectory of ball
			s = (py - px * ratio - ry + rx * ratio) / (lx * ratio - ly);
		}
///*
		if (dBallAbsSpeed > 1.0)
		{
			// look 2 timesteps into future if ball is fast
			if (s > 2.0)
				s = 2.0;
		}
		else
		{

			#ifdef _DEBUG
				TRACE3("Low ball speed %d: speed: %4.2f  dir: %d"
					, m_pPerception->GetSenseTime(), dBallAbsSpeed, (int) ballSpeed.GetDirection());
				TRACE1(" s: %4.2f\n", s);
			#endif

			// just look 1.4 timesteps into future if ball is slow
			if (s > 1.4)
				s = 1.4;
		}
//* /
		// look 2 timesteps into future if ball is fast
		if (s > 2.0)
			s = 2.0;
		
		if (s > 0.0)
			// take the intersection point to evaluate kickability
			// in all other cases the current ball position is taken
			intersection = ballNow + s * ballSpeed;
		else
			// the ball is moving away. Avoid kick faults
			maxKickDistance -= 0.15;
	}
	
	// get the distance to the intersection point
	dMinDist = m_pPerception->GetThisPlayer()->GetDistanceTo(intersection);
		

	if (dMinDist > maxKickDistance)
		// ball is too far
		return 0.0;
	else 
	{
		#ifdef SHOW_VIEWING_VAR
			TRACE2("ballKickable: distance %4.2f  s: %4.2f\n", dMinDist, s);
		#endif
		// the ball is kickable
		return GetHaveSeenBall();
	}
*/  
}

/*-----------------------------------------------------------------------------
true at kick_in or corner_kick
-----------------------------------------------------------------------------*/
double TViewingAgent::GetBallOutside()
{
	if (m_pPerception->GetSide() == LEFT_TO_RIGHT)
	{
		if ((m_pPerception->GetPlayMode() == KICK_IN_L)
			|| (m_pPerception->GetPlayMode() == CORNER_KICK_L))
			// we have kick in
			return 1.0;
	}
	else
	{
		if ((m_pPerception->GetPlayMode() == KICK_IN_R)
			|| (m_pPerception->GetPlayMode() == CORNER_KICK_R))
			// we have kick in
			return 1.0;
	}
	return 0.0;
}

/*-----------------------------------------------------------------------------
true, immediately after a successfull catch
-----------------------------------------------------------------------------*/
double TViewingAgent::GetCatchSuccess()
{
	if (m_pPerception->GetGoalieCaught())
		return 1.0;
	else
		return 0.0;
}

/*-----------------------------------------------------------------------------
true, if the free teammate is in offside
-----------------------------------------------------------------------------*/
double TViewingAgent::GetFreeTeammateIsOffside()
{
	TPlayer*	pTeammate = m_pPerception->GetFreeTeammate();
	
	if (!GetOrdinaryGame() && (pTeammate != NULL))
	{
		// check if the player is offside and if so, do not pass to him
		double	dTeammateX = pTeammate->GetAbsX();
		double	dBallX = m_pPerception->GetBall()->GetAbsX();
		TPlayer*	pOpp = m_pPerception->Get2BackmostOpponent();

		if (m_pPerception->GetSide() == LEFT_TO_RIGHT)
		{
			// check if teammate is in other side and before ball
			if ((dTeammateX > 0) && (dTeammateX > dBallX)
				&& ((pOpp == NULL) || (pOpp->GetAbsX() < dTeammateX)))
			{
				// the teammate is offside
				return 1.0;
			}
		}
		else
		{
			// check if teammate is in other side and before ball
			if ((dTeammateX < 0) && (dTeammateX < dBallX)
				&& ((pOpp == NULL) || (pOpp->GetAbsX() > dTeammateX)))
			{
				// the teammate is offside
				return 1.0;
			}
		}
	}
	return 0.0;
}

/*-----------------------------------------------------------------------------
true, if this agent is nearest to ball and near enough (but maybe farer than ballKickable)
-----------------------------------------------------------------------------*/
double TViewingAgent::GetHaveBall()
{
	TPlayer*		pPlayerAtBall = m_pPerception->GetTeammateAtBall();
	double		ballDist = m_pPerception->GetObjectDistance(m_pPerception->GetBall());

	// is agent nearest to Ball
	if (pPlayerAtBall != NULL)
	{
		// look, if the other agent is nearer to ball
		if (pPlayerAtBall->GetDistanceTo(m_pPerception->GetBall()) < ballDist)
			// the other player is nearer to ball
			return 0.0;
	}

	return GetLinearFuzzyValue(ballDist, 1.0, 9.0, FALSE) * GetHaveSeenBall();
}

/*-----------------------------------------------------------------------------
true, if this agent has enough stamina to dash 100% to the ball
-----------------------------------------------------------------------------*/
double TViewingAgent::GetHaveEnoughStamina()
{
	double	ballDist = m_pPerception->GetObjectDistance(m_pPerception->GetBall());
	int		stamina = m_pPerception->GetStamina();

	return GetLinearFuzzyValue(stamina - ballDist * 60, 700, 1200);
}

/*-----------------------------------------------------------------------------
true, if the ball has been seen within the last 3 seconds
-----------------------------------------------------------------------------*/
double TViewingAgent::GetHaveSeenBall()
{
	// if the ball is not in mind, it has not been seen
	if (!m_pPerception->GetBall()->IsInMind())
		return 0.0;
	
	int deltaTime = m_pPerception->GetSeeTime() - m_pPerception->GetObjectTime(m_pPerception->GetBall());
	if (m_pPerception->GetViewAngle() == 180)
		// we allow longer times if we look around
		return GetLinearFuzzyValue((double)deltaTime, 6.0, (double) MAX_IMAGINE_TIME, FALSE);	
	else
		return GetLinearFuzzyValue((double)deltaTime, 0.0, (double) MAX_IMAGINE_TIME, FALSE);	
}

/*-----------------------------------------------------------------------------
true, if no opponent is close to me
-----------------------------------------------------------------------------*/
double TViewingAgent::GetHaveTime()
{
// see nearOpponent
	return 1.0;
}

/*-----------------------------------------------------------------------------
true if the player has got enough power to dash 
-----------------------------------------------------------------------------*/
double TViewingAgent::GetHighStamina()
{
	return GetLinearFuzzyValue(m_pPerception->GetStamina(), 1500.0, 2000.0);
}

/*-----------------------------------------------------------------------------
true, if the agent knows he is in offside
-----------------------------------------------------------------------------*/
double TViewingAgent::GetInOffside()
{
	 return m_pPerception->GetInOffside();
}

/*-----------------------------------------------------------------------------
true if the agent is in the opponents half
-----------------------------------------------------------------------------*/
double TViewingAgent::GetInOtherHalf()
{
	if (m_pPerception->GetThisPlayer()->IsInOwnHalfOfField())
		return 0.0;
	else
		return 1.0;
}

/*-----------------------------------------------------------------------------
true, if the goalie caught the ball within the last 5 cycles. A goalie can only
catch the ball once each 5 cycles
-----------------------------------------------------------------------------*/
double TViewingAgent::GetJustCaughtBall()
{
	int	lastCatch = m_pPerception->GetSenseTime() - m_pPerception->GetLastCatchTime();
	if (lastCatch < 6)
		return 1.0;
	else
		return 0.0;
}

/*-----------------------------------------------------------------------------
true if the player knows he is not in offside
-----------------------------------------------------------------------------*/
double TViewingAgent::GetKnowNotInOffside()
{
	if (m_pPerception->GetKnowNotInOffside())
		return 1.0;
	else
		return 0.0;
}

/*-----------------------------------------------------------------------------
true if the player has got low power to dash 
-----------------------------------------------------------------------------*/
double TViewingAgent::GetLowStamina()
{
	int	iStart = LOW_STAMINA;
	int	iTime = m_pPerception->GetSenseTime();

	if ((iTime > 5500) && (iTime < 6000))
		// end of second half
		iTime -= 3000;

	if ((iTime > 2800) && (iTime < 3000))
		// end of first or second half
		// full resource usage at end of a half
		iStart = 0;
	else if (iTime > 2500)
		// higher resource usage close to the end of a half
		iStart = 500;
	else
	{
		// if the ball is close, we push harder
		double ballDist = m_pPerception->GetThisPlayer()->GetDistanceTo(*m_pPerception->GetFutureBallPosition(5));
		if (ballDist < 4.0)
			iStart = LOW_STAMINA - (int) (400.0 * GetLinearFuzzyValue(ballDist, 2.0, 4.0, FALSE));
	}

	return GetLinearFuzzyValue(m_pPerception->GetStamina(), iStart, iStart + 300, FALSE);
}

/*-----------------------------------------------------------------------------
true if the player is on turn to say something
-----------------------------------------------------------------------------*/
double TViewingAgent::GetMaySaySomething()
{
	int	time = m_pPerception->GetSenseTime();

	// a player may say something each 22 cycle
	// the players say in order 1 5 9 2 6 10 3 7 11 4 8
	if ((m_pPerception->GetTimeProceeding()) && GetNextSayTime() == time)
		return 1.0;
	
	// we may also say something if time is stopped and player with lower number said sth
	// the players say in order 1 2 3 4 ...
	else if (!m_pPerception->GetTimeProceeding() 
		&& ((m_pPerception->GetLastHearNumber() == m_iNumber - 1)
			||(((m_pPerception->GetLastHearNumber() == 11) && (m_iNumber == 1)))))
		return 1.0;

	else
		return 0.0;
}

/*-----------------------------------------------------------------------------
true if the goalie may wait with kicking away ball in her hands
-----------------------------------------------------------------------------*/
double TViewingAgent::GetMayWait()
{
	// a goalie may wait if the wait time is higher than the current time
	// (set by moveToHomePosition)
	if (m_iMayWaitTime - m_pPerception->GetSenseTime() > 0)
		return 1.0;
	else
		return 0.0;
}

/*-----------------------------------------------------------------------------
true if this player is close to the ball
-----------------------------------------------------------------------------*/
double TViewingAgent::GetNearBall()
{
	return GetLinearFuzzyValue(m_pPerception->GetObjectDistance(m_pPerception->GetBall())
										, 5.0, 15.0, FALSE)
			 * GetHaveSeenBall();
}

/*-----------------------------------------------------------------------------
true if this player is close to the other goal
-----------------------------------------------------------------------------*/
double TViewingAgent::GetNearGoal()
{
	double			goaldistance = m_pPerception->GetObjectDistance(m_pPerception->GetOtherGoal());

	if (goaldistance < 10.0)
		return 1.0;

//	if (goaldistance > 25.0)
//		return 0.0;

	TPlayer*		goalie = m_pPerception->GetOtherGoalie();
	if (goalie != NULL)
	{
		CString		msg;
		// reachable area of goalie
		double		distance = goalie->GetDistance();
		if (distance < 1.0)
			distance = 1.0;

		double delay = 2.0 - 3.0 * GetLinearFuzzyValue(distance, 10.0, 25.0);

		// space the goalie can run. we calculate one cycle delay
		double		radius = (distance / BALL_SPEED_MAX - delay) * PLAYER_SPEED_MAX + CATCHABLE_AREA_L;
		
		int			gamma = (int) (RAD2DEG * atan(radius / distance));
		int			goalieDir = m_pPerception->GetObjectDirection(goalie);
		TDirection	goalieLeft = goalieDir - gamma;
		TDirection	goalieRight = goalieDir + gamma;
		int			dirLeft = m_pPerception->GetObjectDirection(m_pPerception->GetOtherLeftGoalPost());
		int			dirRight = m_pPerception->GetObjectDirection(m_pPerception->GetOtherRightGoalPost());
		m_deltaLeft = goalieLeft - dirLeft;
		m_deltaRight = dirRight - goalieRight;

/*		
		double		gx = goalie->GetAbsX() - m_pPerception->GetThisPlayer()->GetAbsX();
		double		gy = goalie->GetAbsY() - m_pPerception->GetThisPlayer()->GetAbsY();
		TVector		pos;
		int			dirLeft = m_pPerception->GetObjectDirection(m_pPerception->GetOtherLeftGoalPost());
		int			dirRight = m_pPerception->GetObjectDirection(m_pPerception->GetOtherRightGoalPost());
		int			m_deltaRight = 0;
		int			m_deltaLeft = 0;

		CalculateGoalAngle(gx, gy, dirRight, dirLeft, radius, m_deltaLeft, m_deltaRight);

		msg.Format("near Goal: gx:%4.2f gy:%4.2f radius: %4.2f left: %d right: %d", gx, gy, radius, m_deltaLeft, m_deltaRight);
		m_pView->DisplayDebug(msg, KATEGORY_PERCEPTION, DEBUG_LOWPERCEPTION);

*/
		msg.Format("near Goal: radius: %4.2f left:%d right:%d dleft: %d dright: %d gamma:%d goalieDir:%d gx:%4.2f gy:%4.2f"
					, radius, dirLeft, dirRight, (int) m_deltaLeft, (int) m_deltaRight, gamma, goalieDir, goalie->GetAbsX(), goalie->GetAbsY());
		m_pView->DisplayDebug(msg, KATEGORY_PERCEPTION, DEBUG_LOWPERCEPTION);
		
		if (m_deltaLeft > m_deltaRight)
			m_deltaRight = m_deltaLeft;

		if (m_deltaRight > 0)
			return 1.0;
		else
			return GetLinearFuzzyValue(goaldistance, 10.0, 18.0, FALSE);
	}

	m_deltaLeft = -1;
	m_deltaRight = -1;
	return GetLinearFuzzyValue(goaldistance, 15.0, 22.0, FALSE);
}

/*-----------------------------------------------------------------------------
true if this player is close to an opponent
-----------------------------------------------------------------------------*/
double TViewingAgent::GetNearOpponent()
{
	TPlayer*	player = m_pPerception->GetOpponentAtMe();
	if (player == NULL)
		return 0.0;
	
	return GetLinearFuzzyValue(m_pPerception->GetObjectDistance(player)
										, 3.0, 7.0, FALSE);
}

/*-----------------------------------------------------------------------------
true if this player is close to the own goal
-----------------------------------------------------------------------------*/
double TViewingAgent::GetNearOwnGoal()
{
	return GetLinearFuzzyValue(m_pPerception->GetObjectDistance(m_pPerception->GetOwnGoal())
										, 10.0, 30.0, FALSE);
}

/*-----------------------------------------------------------------------------
only false in the case that the opponent scored a goal
-----------------------------------------------------------------------------*/
double TViewingAgent::GetNoOtherGoal()
{
	if (((m_pPerception->GetSide() == LEFT_TO_RIGHT) 
			&& (m_pPerception->GetPlayMode() == GOAL_R)) 
		|| ((m_pPerception->GetSide() == RIGHT_TO_LEFT)
			&& (m_pPerception->GetPlayMode() == GOAL_L)))
		return 0.0;
	else
		return 1.0;
}

/*-----------------------------------------------------------------------------
just for goal offsideTrap
-----------------------------------------------------------------------------*/
double TViewingAgent::GetOffsideTrap()
{
	return 0.0;
}

/*-----------------------------------------------------------------------------
true if an opponent is between this agent and the opponents goal
-----------------------------------------------------------------------------*/
double TViewingAgent::GetOpponentInFront()
{
	// is there a player in the direction to the other goal
	TPlayer* pPlayer = m_pPerception->GetPlayerInWay(m_pPerception->GetObjectDirection(
								m_pPerception->GetOtherGoal()), 15.0);
	if (pPlayer == NULL)
		return 0.0;

	// there is a player
	return GetLinearFuzzyValue(m_pPerception->GetObjectDistance(pPlayer), 6.0, 15.0, FALSE);
}

/*-----------------------------------------------------------------------------
true if playmode is a goalkick for the other team
-----------------------------------------------------------------------------*/
double TViewingAgent::GetPlaymodeGoalkickOther()
{
	if (m_pPerception->GetSide() == LEFT_TO_RIGHT)
	{
		if (m_pPerception->GetPlayMode() == GOAL_KICK_R)
			return 1.0;
	}
	else if (m_pPerception->GetPlayMode() == GOAL_KICK_L) 
		return 1.0;
	
	return 0.0;
}

/*-----------------------------------------------------------------------------
true if playmode is play_on
-----------------------------------------------------------------------------*/
double TViewingAgent::GetPlaymodeOn()
{
	if (m_pPerception->GetPlayMode() == PLAY_ON) 
		return 1.0;
	else
		return 0.0;
}

/*-----------------------------------------------------------------------------
true if a goal has been scored this cycle
-----------------------------------------------------------------------------*/
double TViewingAgent::GetSoccergoal()
{
	if (((m_pPerception->GetSide() == LEFT_TO_RIGHT) 
			&& (m_pPerception->GetPlayMode() == GOAL_L)) 
		|| ((m_pPerception->GetSide() == RIGHT_TO_LEFT)
			&& (m_pPerception->GetPlayMode() == GOAL_R)))
		return 1.0;
	else
		return 0.0;
}

/*-----------------------------------------------------------------------------
true, if this agent is unmarked i.e. the ball can be passed to him
-----------------------------------------------------------------------------*/
double TViewingAgent::GetStandFree()
{
	// opposite to GetNearOpponent
	TPlayer*	player = m_pPerception->GetOpponentAtMe();
	if (player == NULL)
		return 1.0;
	
	return GetLinearFuzzyValue(m_pPerception->GetObjectDistance(player), 3.0, 8.0);
}

/*-----------------------------------------------------------------------------
true if a teammate is behind me in x direction (used for offside trap)
-----------------------------------------------------------------------------*/
double TViewingAgent::GetTeammateBehind()
{
	TPlayer*	pPlayer = m_pPerception->Get2BackmostTeammate();

	if (pPlayer != NULL)
	{
		double	dTeammateX = pPlayer->GetAbsX();
		double	dAgentX = m_pPerception->GetThisPlayer()->GetAbsX();

		if (m_pPerception->GetSide() == RIGHT_TO_LEFT)
			dTeammateX -= dAgentX;
		else
			dTeammateX = dAgentX - dTeammateX;
	
		return GetLinearFuzzyValue(dTeammateX, 4.0, 9.0);
	}
	else
		return 0.0;
}

/*-----------------------------------------------------------------------------
true if the teammate is the closest player to the ball and is close enough
-----------------------------------------------------------------------------*/
double TViewingAgent::GetTeammateHasBall()
{
	TPlayer*	pPlayer = m_pPerception->GetAnyPlayerAtBall();

	if ((pPlayer != NULL) && (pPlayer->IsOwnTeam()))
		return GetLinearFuzzyValue(pPlayer->GetDistanceTo(m_pPerception->GetBall()), 1.0, 9.0, FALSE);
	else
		return 0.0;
}

/*-----------------------------------------------------------------------------
true if a teammate is in front of me in x direction (used for offside prevention)
-----------------------------------------------------------------------------*/
double TViewingAgent::GetTeammateInFront()
{
	TPlayer*	pPlayer = m_pPerception->GetFrontmostTeammate();

	if (pPlayer != NULL)
	{
		double	dTeammateX = pPlayer->GetAbsX();
		double	dAgentX = m_pPerception->GetThisPlayer()->GetAbsX();

		if (m_pPerception->GetSide() == RIGHT_TO_LEFT)
			dTeammateX = dAgentX - dTeammateX;
		else
			dTeammateX -= dAgentX;
	
		return GetLinearFuzzyValue(dTeammateX, 3.0, 5.0);
	}
	else
		return 0.0;
}

/*-----------------------------------------------------------------------------
prevents the goalie from running to the ball if a teammate is between him and the ball.
The goalie has to run to the ball even if he is not the closest teammate to the ball.
This is the case, if an opponent followed by a teammate is running to the goal
-----------------------------------------------------------------------------*/
double TViewingAgent::GetTeammateIsBetweenBall()
{
	// checks, if a teammate is nearer to the ball and between me and the ball
	double	dNearer = GetTeammateIsNearerBall();

	if (dNearer > 0.5)
	{
		// teammate is nearer, so check if he is nearer to the own goal as
		// opponent at ball is
		TPlayer*				ownPlayer = m_pPerception->GetTeammateAtBall();
		TPlayer*				opponent = m_pPerception->GetOpponentAtBall();
		TVisibleObject*	ownGoal = m_pPerception->GetOwnGoal();
		double				deltaDist;

		if ((ownPlayer == NULL) || (opponent == NULL))
			return dNearer;

		deltaDist = opponent->GetDistanceTo(ownGoal) - ownPlayer->GetDistanceTo(ownGoal);
		return GetLinearFuzzyValue(deltaDist, 1.0, 3.0) * dNearer;
	}
	else
		return dNearer;
}

/*-----------------------------------------------------------------------------
true if one of the teammates is unmarked
-----------------------------------------------------------------------------*/
double TViewingAgent::GetTeammateIsFree()
{
	TPlayer*		pFreeTeammate = m_pPerception->GetFreeTeammate();

	if (pFreeTeammate == NULL)
		return 0.0;
	else
	{
//		return pFreeTeammate->GetFreePassDegree();
		int		age = m_pPerception->GetSeeTime() - pFreeTeammate->GetTime();
		return	pFreeTeammate->GetFreePassDegree() * GetLinearFuzzyValue(age, 3, 8, FALSE);
	}
}


/*-----------------------------------------------------------------------------
true if a teammate is closer to the ball than this player
-----------------------------------------------------------------------------*/
double TViewingAgent::GetTeammateIsNearerBall()
{
	int			intersectionTime = 0;
	int			iClosest		= -1;
	int			i2Closest		= -1;
	int			iClosestTime	= 100;
	int			i2ClosestTime	= 200;
	TPlayer*	player;
	TPlayer*	ownPlayer		= NULL;
	TVector		ballInFuture;
	CString		msg;

	// look all players if they are closer
	for (int i = 0; i < PLAYERS_PER_TEAM; i++)
	{
		player = m_pPerception->GetOwnPlayer(i);
		if (player != NULL)
		{
			if (player->IsThisAgent())
				player = m_pPerception->GetThisPlayer();
			intersectionTime = GetIntersectionTime(player, 100);
			if (intersectionTime < iClosestTime)
			{
				i2ClosestTime = iClosestTime;
				iClosestTime = intersectionTime;
				i2Closest = iClosest;
				iClosest = i;
			}
			else if (intersectionTime < i2ClosestTime)
			{
				i2ClosestTime = intersectionTime;
				i2Closest = i;
			}
		}
	}

	msg.Format("teammate closer: iClosest:%d time:%d"
					, iClosest, iClosestTime);
	m_pView->DisplayDebug(msg, KATEGORY_PERCEPTION, DEBUG_LOWPERCEPTION);

	if (iClosest + 1 == m_pPerception->GetThisPlayer()->GetNum())
		// this player is closest to ball
		return 0.0;

	if ((iClosestTime < 0) || (iClosestTime >= MAX_FUTURE_POSITIONS))
		// no one is reaching the ball within 20 cycles
		return OldGetTeammateIsNearerBall();

	ballInFuture = *m_pPerception->GetFutureBallPosition(iClosestTime);

	// if the ball is close to the own goal, two players should attack
	if (m_pPerception->GetOwnGoal()->GetDistanceTo(ballInFuture) < 30)
	{
		// but only if there is an opponent near the ball
		if (m_pPerception->GetOpponentNearBall(5.0))
			ownPlayer = m_pPerception->Get2TeammateAtBall();
	}

	if (ownPlayer == NULL)
		ownPlayer = m_pPerception->GetTeammateAtBall();

	if ((ownPlayer == NULL) || (!ownPlayer->IsInMind()))
		// if there is no one playing with me
		return 0.0;
	
	double deltaDist = m_pPerception->GetThisPlayer()->GetDistanceTo(ballInFuture)
					- ownPlayer->GetDistanceTo(ballInFuture);

	double value = GetLinearFuzzyValue(deltaDist, 0.0, 1.0);

/*	msg.Format("teammate closer: deltadist: %4.2f who: %d value:%4.2f"
					, deltaDist, ownPlayer->GetNum(), value);
	m_pView->DisplayDebug(msg, KATEGORY_PERCEPTION, DEBUG_LOWPERCEPTION);
*/
	return value;
}

/*-----------------------------------------------------------------------------
true if a teammate is closer to the ball than this player
-----------------------------------------------------------------------------*/
double TViewingAgent::OldGetTeammateIsNearerBall()
{
	// that player is nearer ball, that will be faster at the balls future position
	CString msg;
	TThisPlayer* pAgent = m_pPerception->GetThisPlayer();
	TPlayer*		ownPlayer = NULL;
	double		deltaDist;
	TVector		ballInFuture;
	TBall*		pBall = m_pPerception->GetBall();
	TVector		tBallSpeed = pBall->GetRelSpeed(pAgent->GetPosition(), pAgent->GetAbsDir());
	int			ballRelDir = tBallSpeed.GetDirection();
	double		value;

	// if the ball is moving away, look farer into future
	if ((ballRelDir > -90) && (ballRelDir < 90))
		// ball moves away
		ballInFuture = *m_pPerception->GetFutureBallPosition(14);
	else
	{
		// ball approaches 
		// look for closest approch
		double	dMinDist = 1000.0;
		int		iMinIndex = 0;
		double	distance;

		for (int i = 0; i < 18; i++)
		{
			ballInFuture = *m_pPerception->GetFutureBallPosition(i);
			distance = pAgent->GetDistanceTo(ballInFuture);
			if (distance < dMinDist)
			{
				dMinDist = distance;
				iMinIndex = i;
			}
		}
		ballInFuture = *m_pPerception->GetFutureBallPosition(iMinIndex);
	}

	// if the ball is close to the own goal, two players should attack
	if (m_pPerception->GetOwnGoal()->GetDistanceTo(ballInFuture) < 30)
	{
		// but only if there is an opponent near the ball
		if (m_pPerception->GetOpponentNearBall(5.0))
			ownPlayer = m_pPerception->Get2TeammateAtBall();
	}

	if (ownPlayer == NULL)
		ownPlayer = m_pPerception->GetTeammateAtBall();

	if ((ownPlayer == NULL) || (!ownPlayer->IsInMind()))
		// if there is no one playing with me
		return 0.0;
	
	deltaDist = pAgent->GetDistanceTo(ballInFuture)
					- ownPlayer->GetDistanceTo(ballInFuture);

	value = GetLinearFuzzyValue(deltaDist, 0.0, 1.0);

/*	msg.Format("teammate closer: deltadist: %4.2f who: %d value:%4.2f"
					, deltaDist, ownPlayer->GetNum(), value);
	m_pView->DisplayDebug(msg, KATEGORY_PERCEPTION, DEBUG_LOWPERCEPTION);
*/	return value;
}


/*-----------------------------------------------------------------------------
true if the accessible area to the other goal is big
-----------------------------------------------------------------------------*/
double TViewingAgent::GetWayToGoalFree()
{
	int		goalDir = m_pPerception->GetObjectDirection(m_pPerception->GetOtherGoal());
	double	dist = m_pPerception->GetAccessDistance(goalDir);
	
	return GetLinearFuzzyValue(dist, 3.0, 7.0);
}

/*-----------------------------------------------------------------------------
Function: void TViewingAgent::CreateResources()
Date    : 22.02.99
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: adds the perception functions to the behavior networks function factory
-----------------------------------------------------------------------------*/
void TViewingAgent::CreateResources()
{
	m_pBehaviorNetwork->AddResource("leg", this, (PCallBackPerception) GetLegsAvailable, 1);
	m_pBehaviorNetwork->AddResource("mouth", this, (PCallBackPerception) GetMouthsAvailable, 1);
	m_pBehaviorNetwork->AddResource("neck", this, (PCallBackPerception) GetNecksAvailable, 1);
	m_pBehaviorNetwork->AddResource("stamina", this, (PCallBackPerception) GetStaminaAvailable, MAX_STAMINA);
}	


double TViewingAgent::GetLegsAvailable()
{
	return 1.0;
}

double TViewingAgent::GetMouthsAvailable()
{
	return 1.0;
}

double TViewingAgent::GetNecksAvailable()
{
	return 1.0;
}

double TViewingAgent::GetStaminaAvailable()
{
	return m_pPerception->GetStamina();
}


// inserted for compatibility with agent3

double TViewingAgent::GetOpponentIsNearerToBall()
{
	TPlayer*	pPlayer = m_pPerception->GetOpponentAtBall();
	double	deltaDist;

	if (pPlayer == NULL)
		// if there is no opponent near the ball
		return 0.0;
	
	deltaDist = m_pPerception->GetObjectDistance(m_pPerception->GetBall())
					- pPlayer->GetDistanceTo(m_pPerception->GetBall());
	return GetLinearFuzzyValue(deltaDist, 0.0, 1.0);
}

double TViewingAgent::GetTeammateIsNearerOtherGoal()
{
	TPlayer*	ownPlayer = m_pPerception->GetTeammateAtMe();
	double	deltaDist;

	if ((ownPlayer == NULL) || (!ownPlayer->IsInMind()))
		// if there is no one playing with me
		return 0.0;
	
	deltaDist = m_pPerception->GetObjectDistance(m_pPerception->GetOtherGoal())
					- ownPlayer->GetDistanceTo(m_pPerception->GetOtherGoal());
	return GetLinearFuzzyValue(deltaDist, 0.0, 1.0);
}

double TViewingAgent::GetOpponentIsNearerToOwnGoal()
{
	TPlayer*	pPlayer = m_pPerception->GetOpponentAtBall();
	double	deltaDist;

	if (pPlayer == NULL)
		// if there is no opponent near the ball
		return 0.0;
	
	deltaDist = m_pPerception->GetObjectDistance(m_pPerception->GetOwnGoal())
					- pPlayer->GetDistanceTo(m_pPerception->GetOwnGoal());
	return GetLinearFuzzyValue(deltaDist, 0.0, 1.0);
}

double TViewingAgent::GetTrue()
{
	return 1.0;
}





void TViewingAgent::CalculateGoalAngle(double gx, double gy, int wr, int wl, double r, 
									 int& m_deltaLeft, int& m_deltaRight)
{
    double a = 4*r*r -4*gx*gx; //4 * radius^2 form gaolkeeper

    if ((a > D_ZERO_N) && (a < D_ZERO_P))
		a = 0.001;

    double b = 8*gx*gy;
    double c = -4*gy*gy +4*r*r;
    double root = b*b - 4*a*c;
    if (root < 0.0)
	{
        root = 0.0;
    }
    
	double k1 = (-b + sqrt(root)) / (2*a);
    double k2 = (-b - sqrt(root)) / (2*a);
    k1 = RAD2DEG * atan(k1);
    k2 = RAD2DEG * atan(k2);
    if (gy > 0 )
	{
        if (k1 < 0) 
            k1 = 180 + k1;
        if (k2 < 0)
            k2 = 180 + k2;
    } 
	else 
	{
        if (k1 > 0) 
            k1 = -180 + k1;
        if (k2 > 0) 
            k2 = -180 + k2;
    }                
    m_deltaRight = (int) k1 - wr;
    m_deltaLeft = wl - (int) k2;
}
